from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from xproject.xspider.xitems.xitem import Item


class Model(ABC):
    @classmethod
    @abstractmethod
    def columns(cls) -> list[str]:
        ...

    @classmethod
    def save(cls, data: dict[str, Any] | Item) -> bool:
        ...



import logging
import os
import sys
from logging import handlers
from typing import Any, Literal, Final, TypeVar, cast

import colorlog

FRAMEWORK: Final[int] = 55
SUCCESS: Final[int] = 25

logging.addLevelName(FRAMEWORK, "FRAMEWORK")
logging.addLevelName(SUCCESS, "SUCCESS")

LoggerLevel = Literal["DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL", "FRAMEWORK"]
LoggerLevelT = TypeVar("LoggerLevelT", bound=LoggerLevel)


class Logger(logging.Logger):
    def framework(
            self,
            msg: str,
            *args: Any,
            level: LoggerLevelT = "DEBUG",
            **kwargs: Any
    ) -> None:
        if self.isEnabledFor(FRAMEWORK):
            msg = f"[{level}] {msg}"
            self._log(FRAMEWORK, msg, args, **kwargs, stacklevel=2)

    def success(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(SUCCESS):
            self._log(SUCCESS, msg, args, **kwargs, stacklevel=2)


logging.setLoggerClass(Logger)


class RotatingFileHandler(handlers.RotatingFileHandler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._zero_padding: Final[int] = len(str(self.backupCount)) if self.backupCount > 0 else 0

    def doRollover(self) -> None:
        if self.stream:
            self.stream.close()
            self.stream = None  # noqa

        if self.backupCount > 0:
            oldest = self._format_backup_filename(self.backupCount)
            if os.path.exists(oldest):
                os.remove(oldest)

            for i in range(self.backupCount - 1, 0, -1):
                sfn = self._format_backup_filename(i)
                dfn = self._format_backup_filename(i + 1)
                if os.path.exists(sfn):
                    os.rename(sfn, dfn)

            dfn = self._format_backup_filename(1)
            if os.path.exists(self.baseFilename):
                os.rename(self.baseFilename, dfn)

        self.mode = "w"
        self.stream = self._open()

    def _format_backup_filename(self, index: int) -> str:
        base, ext = os.path.splitext(self.baseFilename)
        padded = str(index).zfill(self._zero_padding)
        return f"{base}.{padded}{ext}"


def _set_console_handler(
        _logger: Logger,
        console_level: str | int, console_fmt: str
) -> Logger:
    console_handler = logging.StreamHandler()
    console_handler.setLevel(console_level)
    green = {
        "DEBUG": "green",
        "INFO": "green",
        "SUCCESS": "green",
        "WARNING": "green",
        "ERROR": "green",
        "CRITICAL": "green",
        "FRAMEWORK": "green",
    }
    bold_cyan = {
        "DEBUG": "bold_cyan",
        "INFO": "bold_cyan",
        "SUCCESS": "bold_cyan",
        "WARNING": "bold_cyan",
        "ERROR": "bold_cyan",
        "CRITICAL": "bold_cyan",
        "FRAMEWORK": "bold_cyan",
    }
    log_colors = {
        "DEBUG": "light_blue",
        "INFO": "light_white",
        "SUCCESS": "light_green",
        "WARNING": "light_yellow",
        "ERROR": "light_red",
        "CRITICAL": "bg_red,light_white",
        "FRAMEWORK": "light_red,bold",
    }
    secondary_log_colors = dict(
        asctime=green,
        name=bold_cyan,
        levelname=log_colors,
        process=bold_cyan,
        processName=bold_cyan,
        thread=bold_cyan,
        threadName=bold_cyan,
        pathname=bold_cyan,
        funcName=bold_cyan,
        lineno=bold_cyan,
        message=log_colors,
    )
    console_formatter = colorlog.ColoredFormatter(
        console_fmt,
        secondary_log_colors=secondary_log_colors,
    )
    console_handler.setFormatter(console_formatter)
    _logger.addHandler(console_handler)
    return _logger


def _set_file_handler(
        _logger: Logger,
        file_level: str | int, file_fmt: str,
        file_path: str, file_mode: str, file_max_bytes: int, file_backup_count: int, file_encoding: str,
) -> Logger:
    file_handler = RotatingFileHandler(
        file_path,
        mode=file_mode, maxBytes=file_max_bytes, backupCount=file_backup_count, encoding=file_encoding
    )
    file_handler.setLevel(file_level)
    file_formatter = logging.Formatter(file_fmt)
    file_handler.setFormatter(file_formatter)
    _logger.addHandler(file_handler)
    return _logger


def get_logger(
        name: str | None = None,
        level: str = "DEBUG",
        to_console: bool = True,
        console_level: str = "DEBUG",
        console_fmt: str = (
                "%(asctime_log_color)s%(asctime)s %(reset)s| "
                "%(name_log_color)s%(name)s %(reset)s| "
                "%(levelname_log_color)s%(levelname)s %(reset)s| "
                # "%(process_log_color)sProcess: %(process)d %(processName_log_color)s(%(processName)s) %(reset)s| "
                # "%(thread_log_color)sThread: %(thread)d %(threadName_log_color)s(%(threadName)s) %(reset)s| "
                # "%(pathname_log_color)s%(pathname)s %(reset)s| "
                "%(funcName_log_color)s%(funcName)s:%(lineno)d %(reset)s- "
                "%(message_log_color)s%(message)s"
        ),
        to_file: bool = False,
        file_level: str = "DEBUG",
        file_path: str | None = None,
        file_mode: str = "a",
        file_max_bytes: int = 10 * 1024 * 1024,
        file_backup_count: int = 20,
        file_encoding: str = "utf8",
        file_fmt: str = (
                "%(asctime)s | "
                "%(name)s | "
                "%(levelname)s | "
                # "Process: %(process)d (%(processName)s) | "
                # "Thread: %(thread)d (%(threadName)s) | "
                # "%(pathname)s | "
                "%(funcName)s:%(lineno)d - "
                "%(message)s"
        ),
) -> Logger:
    _path = os.path.abspath(sys.argv[0])
    _name = os.path.basename(_path)
    _prefix = os.path.splitext(_name)[0]
    _file_name = _prefix + ".log"
    _file_dir = os.path.dirname(_path)
    _file_path = os.path.join(_file_dir, _file_name)

    if name is None:
        name = _prefix
    _logger = cast(Logger, logging.getLogger(name))
    _logger.setLevel(level)

    if to_console:
        console_handler_exists = any(isinstance(handler, logging.StreamHandler) for handler in _logger.handlers)
        if not console_handler_exists:
            _set_console_handler(_logger,
                                 console_level=console_level,
                                 console_fmt=console_fmt)
    else:
        for handler in _logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                _logger.removeHandler(handler)

    if to_file:
        file_handler_exists = any(isinstance(handler, logging.FileHandler) for handler in _logger.handlers)
        if not file_handler_exists:
            if file_path is None:
                file_path = _file_path
            _set_file_handler(_logger,
                              file_level=file_level,
                              file_path=file_path, file_mode=file_mode, file_max_bytes=file_max_bytes,
                              file_backup_count=file_backup_count, file_encoding=file_encoding,
                              file_fmt=file_fmt)
    else:
        for handler in _logger.handlers:
            if isinstance(handler, logging.FileHandler):
                _logger.removeHandler(handler)

    return _logger


logger = get_logger()

__all__ = [
    "LoggerLevel",
    "Logger",
    "get_logger",
    "logger"
]

if __name__ == '__main__':
    logger.debug("debug message")
    logger.info("info message")
    logger.success("success message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")
    logger.framework("framework message")



import socket
from typing import Literal

import httpx


def check_port_occupancy(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        port_occupied = s.connect_ex(("127.0.0.1", port)) == 0
        return port_occupied


def get_host(host_type: Literal["local", "public"] = "local") -> str:
    if host_type == "local":
        return socket.gethostbyname(socket.gethostname())
    elif host_type == "public":
        return httpx.get("https://api.ipify.org", timeout=60).text
    else:
        raise ValueError(
            f"Invalid type for 'host_type': "
            f"Expected `Literal[\"local\", \"public\"]`, "
            f"but got {type(host_type).__name__!r} (value: {host_type!r})"
        )


if __name__ == '__main__':
    print(get_host())
    print(get_host("public"))



from collections.abc import MutableMapping, Iterator
from copy import deepcopy
from pprint import pformat
from typing import Any, Self

from xproject.xtypes import ViewClasses, create_view_classes

Data = dict[str, Any]


class MutableMappingMixin(MutableMapping):
    def __init__(self) -> None:
        self._data: Data = dict()

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}\n{pformat(self._data)}\n>"

    __str__ = __repr__

    def to_dict(self) -> dict[str, Any]:
        return dict(self)

    def copy(self) -> Self:
        return deepcopy(self)

    @classmethod
    @property
    def view_classes(cls) -> ViewClasses:  # noqa
        return create_view_classes(cls.__name__)

    def items(self) -> ViewClasses.ItemsView:
        return self.view_classes.ItemsView(self)

    def keys(self) -> ViewClasses.KeysView:
        return self.view_classes.KeysView(self)

    def values(self) -> ViewClasses.ValuesView:
        return self.view_classes.ValuesView(self)



import base64
import binascii
from collections.abc import Buffer
from typing import Literal, cast

from Crypto.Cipher import AES as AES_
from Crypto.Util.Padding import pad


def zero_pad(data_bytes: bytes, block_size: int = 16) -> bytes:
    pad_len = (block_size - (len(data_bytes) % block_size)) % block_size
    return data_bytes + b"\x00" * pad_len


def zero_unpad(data_bytes: bytes) -> bytes:
    return data_bytes.rstrip(b"\x00")


def string_to_hex(string: str) -> str:
    """
    >>> string_to_hex("hi")
    '\\\\x68\\\\x69'

    """
    return ''.join(f'\\x{ord(s):02x}' for s in string)


def hex_to_string(hex_string: str) -> str:
    """
    >>> hex_to_string("\\\\x68\\\\x69")
    'hi'

    """
    hex_pairs = [hex_string[i:i + 4] for i in range(0, len(hex_string), 4)]
    return "".join(chr(int(pair[2:], 16)) for pair in hex_pairs if pair.startswith("\\x"))


class AES:
    MODE_ECB = AES_.MODE_ECB
    MODE_CBC = AES_.MODE_CBC

    block_size = AES_.block_size

    new = AES_.new

    @staticmethod
    def encrypt(
            data: str,
            key: str,
            iv: str | None = None,
            mode: Literal[1, 2] = MODE_ECB,
            style: Literal["pkcs7", "x923", "iso7816"] = "pkcs7",
            fmt: Literal["base64", "hex"] = "base64"
    ) -> str:
        """
        default: AES ECB PKCS7Padding

        """
        data_bytes: bytes = data.encode()
        key_bytes: bytes = key.encode()
        if iv is not None:
            iv_bytes = iv.encode()
        else:
            iv_bytes = None

        if mode == AES.MODE_ECB:
            cipher = AES.new(key_bytes, mode)
        else:
            cipher = AES.new(key_bytes, mode, iv_bytes)

        padded_data_bytes: bytes = pad(data_bytes, AES.block_size, style)
        encrypted_data_bytes: bytes = cipher.encrypt(padded_data_bytes)
        encrypted_data_buffer: Buffer = cast(Buffer, encrypted_data_bytes)

        if fmt == "base64":
            encrypted_data: str = base64.b64encode(encrypted_data_buffer).decode()
        elif fmt == "hex":
            encrypted_data: str = binascii.hexlify(encrypted_data_buffer).decode()
        else:
            raise ValueError(
                f"Invalid type for 'fmt': "
                f"Expected ` Literal[\"base64\", \"hex\"] `, "
                f"but got {type(fmt).__name__!r} (value: {fmt!r})"
            )
        return encrypted_data


__all__ = [
    "string_to_hex", "hex_to_string",
    "AES",
]

if __name__ == '__main__':
    print(AES.encrypt("123456", "1234567890123456"))
    print(AES.encrypt("123456", "1234567890123456", "1234567890123456", mode=AES.MODE_CBC))
    print(AES.encrypt("123456", "1234567890123456", "1234567890123456", mode=AES.MODE_ECB, style="x923"))



from typing import Any, Self

from redis import Redis

from xproject.xdbs.xdb import DB
from xproject.xurl import get_furl_obj


class RedisDB(DB):
    def __init__(
            self,
            *args: Any,
            host: str = "localhost",
            port: int = 6379,
            password: str | None = None,
            dbname: int = 0,
            **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        self._host = host
        self._port = port
        self._password = password
        self._dbname = dbname

        self._redis: Redis | None = None

    def open(self) -> None:
        self._redis = Redis(
            host=self._host,
            port=self._port,
            db=self._dbname,
            password=self._password,
            encoding="utf-8",
            decode_responses=True
        )

    def close(self) -> None:
        self._redis.close()
        self._redis = None

    @property
    def redis(self) -> Redis:
        return self._redis

    @classmethod
    def from_uri(cls, uri: str) -> Self:
        furl = get_furl_obj(uri)
        kwargs = dict()
        kwargs["host"] = furl.host
        kwargs["port"] = furl.port
        kwargs["password"] = furl.password
        kwargs["dbname"] = furl.path.segments[0]
        return super().from_uri(**kwargs)



from abc import ABC
from typing import Any, Self


class ObjectMixin(ABC):
    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        return super().__new__(cls)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._is_closed: bool | None = None

    @classmethod
    def create_instance(cls, *args: Any, **kwargs: Any) -> Self:
        return cls(*args, **kwargs)

    @property
    def is_closed(self) -> bool:
        return True if self._is_closed is True or self._is_closed is None else False



from __future__ import annotations

import importlib
import lzma
import os
import pprint
import random
import re
import shutil
import sys
import time
from dataclasses import dataclass
from functools import wraps
from types import TracebackType
from typing import TYPE_CHECKING, Callable, ParamSpec, TypeVar

from xproject.xcommand import execute_cmd_code_by_subprocess_popen, execute_cmd_code_by_subprocess_run
from xproject.xlogger import get_logger, Logger
from xproject.xurl import url_to_file_path, get_furl_obj

if TYPE_CHECKING:
    import frida as f
    import _frida as _f

    Device = _f.Device
    Session = _f.Session
    Script = _f.Script
    Application = _f.Application
    Process = _f.Process
else:
    Device = object
    Session = object
    Script = object
    Application = object
    Process = object


@dataclass(frozen=True)
class FridaData:
    frida_server_xz_url: str
    frida_server_version: str
    pc_frida_server_xz_file_path: str
    pc_frida_server_file_path: str
    mobile_frida_server_file_path: str
    mobile_frida_server_process_name: str


P = ParamSpec("P")
R = TypeVar("R")


def frida_log(func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        self: Frida = args[0]
        logger = getattr(self, "logger")
        if logger is not None:
            logger.debug(f"Calling {func.__qualname__}...")
        result = func(*args, **kwargs)
        if logger is not None:
            logger.debug(f"Called {func.__qualname__}.")
        return result

    return wrapper


class Frida:
    def __init__(
            self,
            frida_server_xz_url: str | None = None,
            device_id: str | None = None,
            logger: Logger | None = None,
            timeout: int = 5,
            encoding: str | None = None,
            with_use_open_log: bool = True,
            with_use_init: bool = True
    ) -> None:
        """
        https://github.com/frida/frida/releases
        """
        if not frida_server_xz_url:
            frida_server_xz_url = "https://github.com/frida/frida/releases/download/16.0.0/frida-server-16.0.0-android-arm64.xz"
        self.frida_server_xz_url = frida_server_xz_url

        self.device_id = device_id

        self.logger: Logger | None = None
        if logger is None:
            logger = get_logger()
        else:
            self.logger = logger
        self._logger = logger

        self.timeout = timeout

        self.encoding = encoding

        self.with_use_open_log = with_use_open_log
        self.with_use_init = with_use_init

        self.frida_data = self._create_frida_data()
        assert self.frida_data is not None

        self.device: Device | None = None
        self.session: Session | None = None
        self.script: Script | None = None

    def open_log(self) -> None:
        if self.logger is None:
            self.logger = self._logger

    def close_log(self) -> None:
        if self.logger is not None:
            self.logger = None

    @frida_log
    def _create_frida_data(self) -> FridaData | None:
        try:
            segments = get_furl_obj(self.frida_server_xz_url).path.segments
            assert len(segments) == 6
            assert segments[0] == "frida"
            assert segments[1] == "frida"
            assert segments[2] == "releases"
            assert segments[3] == "download"
            frida_server_version = segments[4]

            from xproject import get_file_assets_dir_path
            assets_dir_path = get_file_assets_dir_path(__file__)
            mobile_dir_path = "/data/local/tmp/"

            frida_server_xz_file_name = os.path.basename(self.frida_server_xz_url)
            pc_frida_server_xz_file_path = os.path.join(assets_dir_path, frida_server_xz_file_name)

            frida_server_file_name = re.sub(r"\.xz$", "", frida_server_xz_file_name, re.DOTALL)
            pc_frida_server_file_path = os.path.join(assets_dir_path, frida_server_file_name)

            mobile_frida_server_file_path = (
                mobile_dir_path + frida_server_file_name
                if mobile_dir_path.endswith("/") else
                mobile_dir_path + "/" + frida_server_file_name
            )
            mobile_frida_server_process_name = frida_server_file_name

            return FridaData(
                self.frida_server_xz_url,
                frida_server_version,
                pc_frida_server_xz_file_path,
                pc_frida_server_file_path,
                mobile_frida_server_file_path,
                mobile_frida_server_process_name
            )
        except:  # noqa
            return None

    @frida_log
    def _download_frida_server_xz(self) -> bool:
        return url_to_file_path(
            self.frida_data.frida_server_xz_url,
            file_path=self.frida_data.pc_frida_server_xz_file_path
        ) == self.frida_data.pc_frida_server_xz_file_path

    @frida_log
    def _decompress_frida_server_xz(self) -> bool:
        with (
            lzma.open(self.frida_data.pc_frida_server_xz_file_path, "rb") as f_in,
            open(self.frida_data.pc_frida_server_file_path, "wb") as f_out
        ):
            shutil.copyfileobj(f_in, f_out)  # type: ignore
            return True

    @frida_log
    def _push_and_chmod_frida_server(self) -> bool:
        text = execute_cmd_code_by_subprocess_run(
            f"adb shell test -e {self.frida_data.mobile_frida_server_file_path} && echo 1 || echo 0",
            encoding=self.encoding,
            logger=self.logger
        ).text
        if text == "0":
            return execute_cmd_code_by_subprocess_run(
                f"adb push {self.frida_data.pc_frida_server_file_path} "
                f"{self.frida_data.mobile_frida_server_file_path}",
                encoding=self.encoding,
                logger=self.logger
            ).result.returncode == 0

        text = execute_cmd_code_by_subprocess_run(
            f"adb shell test -x {self.frida_data.mobile_frida_server_file_path} && echo 1 || echo 0"
        ).text
        if text == "0":
            return execute_cmd_code_by_subprocess_run(
                f"adb shell chmod 755 {self.frida_data.mobile_frida_server_file_path}",
                encoding=self.encoding,
                logger=self.logger
            ).result.returncode == 0

        return True

    @frida_log
    def _pip_install_frida(self) -> bool:
        text = execute_cmd_code_by_subprocess_popen(
            "pip show frida",
            encoding=self.encoding,
            logger=self.logger
        ).text

        if text == "WARNING: Package(s) not found: frida":
            return execute_cmd_code_by_subprocess_popen(
                f"pip install frida=={self.frida_data.frida_server_version}",
                encoding=self.encoding,
                logger=self.logger
            ).returncode == 0

        else:
            if f"Version: {self.frida_data.frida_server_version}" not in text:
                return execute_cmd_code_by_subprocess_popen(
                    f"pip install --upgrade --force-reinstall frida=={self.frida_data.frida_server_version}",
                    encoding=self.encoding,
                    logger=self.logger
                ).returncode == 0

        return True

    @frida_log
    def _pip_install_frida_tools(self) -> bool:
        text = execute_cmd_code_by_subprocess_popen(
            "pip show frida-tools",
            encoding=self.encoding,
            logger=self.logger
        ).text
        if text == "WARNING: Package(s) not found: frida-tools":
            return execute_cmd_code_by_subprocess_popen(
                "pip install frida-tools",
                encoding=self.encoding,
                logger=self.logger
            ).returncode == 0
        return True

    @frida_log
    def _pip_install_frida_and_frida_tools(self) -> bool:
        if not self._pip_install_frida():
            return False

        if not self._pip_install_frida_tools():
            return False

        return True

    @frida_log
    def _import_frida(self) -> bool:
        if not self._pip_install_frida():
            return False
        globals()["f"] = importlib.import_module("frida")

        if not self._pip_install_frida_tools():
            return False
        globals()["_f"] = importlib.import_module("_frida")
        return True

    @frida_log
    def init(self):
        assert self._download_frida_server_xz() is True

        assert self._decompress_frida_server_xz() is True

        assert self._push_and_chmod_frida_server() is True

    @frida_log
    def frida_server_is_running(self) -> bool:
        return execute_cmd_code_by_subprocess_popen(
            f"adb shell pidof {self.frida_data.mobile_frida_server_process_name}",
            encoding=self.encoding,
            logger=self.logger
        ).text != ""

    @frida_log
    def start_frida_server(self) -> bool:
        if not self.frida_server_is_running():
            return execute_cmd_code_by_subprocess_run(
                f'''adb shell su -c "{self.frida_data.mobile_frida_server_file_path} >/dev/null 2>&1 &"''',
                encoding=self.encoding,
                logger=self.logger
            ).text == ""

        return True

    @frida_log
    def stop_frida_server(self) -> bool:
        if self.frida_server_is_running():
            return execute_cmd_code_by_subprocess_popen(
                f'''adb shell su -c "kill -9 $(pidof {self.frida_data.mobile_frida_server_process_name})"''',
                encoding=self.encoding,
                logger=self.logger
            ).returncode == 0

        return True

    @frida_log
    def init_device(self) -> None:
        if not self.frida_server_is_running():
            return

        self._import_frida()

        self.device = f.get_device(self.device_id) if self.device_id else f.get_usb_device(self.timeout)

    @frida_log
    def detach_session(self) -> None:
        if self.session is not None:
            self.session.detach()
            self.session = None

    @frida_log
    def spawn(self, package_names: list[str] | str) -> None:
        self.detach_session()

        if isinstance(package_names, str):
            package_names = [package_names]

        if not package_names:
            return

        pid = self.device.spawn(package_names)  # noqa
        self.session = self.device.attach(pid)
        self.device.resume(pid)
        if self.logger is not None:
            self.logger.info(f"Spawned package_names, package_names: {package_names}")

        seconds = random.uniform(1, 3)
        if self.logger is not None:
            self.logger.info(f"Random sleep for {seconds:.2f} seconds")

        time.sleep(seconds)

    @frida_log
    def get_frontmost_application(self) -> Application | None:
        return self.device.get_frontmost_application()

    @frida_log
    def attach(self, pid_or_process_name: int | str | None = None) -> None:
        self.detach_session()

        if pid_or_process_name is None:
            application = self.get_frontmost_application()
            pid = application.pid
            pid_or_process_name = pid

        if not pid_or_process_name:
            return

        self.session = self.device.attach(pid_or_process_name)
        if self.logger is not None:
            self.logger.info(f"Attached pid_or_process_name, pid_or_process_name: {pid_or_process_name}")

    @frida_log
    def unload_script(self) -> None:
        if self.script is not None:
            self.script.unload()
            self.script = None

    @frida_log
    def load_script(self, js_file_path_or_js_code: str, on_message=None) -> None:
        self.unload_script()

        if not js_file_path_or_js_code:
            return

        if os.path.isfile(js_file_path_or_js_code) and os.path.exists(js_file_path_or_js_code):
            with open(js_file_path_or_js_code, "r", encoding="utf-8") as file:
                js_code = file.read()
        else:
            js_code = js_file_path_or_js_code

        self.script = self.session.create_script(js_code)

        def default_message(message, data):
            if message["type"] == "send":
                text = f"[*] {message['payload']}"
                if self.logger is not None:
                    self.logger.success(text)
                else:
                    print(text)
            else:
                text = f"[*] {pprint.pformat(message)}"
                if self.logger is not None:
                    self.logger.success(text)
                else:
                    print(text)

        self.script.on("message", on_message or default_message)
        self.script.load()
        if self.logger is not None:
            self.logger.info(f"Script loaded, js_file_path_or_js_code: {js_file_path_or_js_code}")

    @frida_log
    def listen(
            self,
            use_unload_script: bool = True,
            use_detach_session: bool = True
    ) -> None:
        try:
            for line in sys.stdin:
                if line.strip().lower() in ("exit", "quit"):
                    break
        except KeyboardInterrupt:
            if self.logger is not None:
                self.logger.warning(f"Interrupted by user, cleaning up")
        finally:
            if use_unload_script:
                self.unload_script()
            if use_detach_session:
                self.detach_session()

    def __enter__(self) -> Frida:
        if self.with_use_open_log:
            self.open_log()
        if self.with_use_init:
            self.init()
        self.start_frida_server()
        self.init_device()
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None
    ) -> None:
        self.stop_frida_server()

    def get_processes(self) -> list[Process]:
        return self.device.enumerate_processes()


if __name__ == '__main__':
    def main1():
        frida = Frida()
        frida.open_log()
        frida.init()
        frida.start_frida_server()
        frida.init_device()

        frida.spawn("com.example.networkapplication")
        # frida.attach()

        # language=JavaScript
        js_code = """
                  Java.perform(function () {
                    let MainActivity = Java.use("com.example.networkapplication.MainActivity");
                    MainActivity["s"].implementation = function (message) {
                      console.log(`MainActivity.s is called: message=${message}`);
                      this["s"](message);
                    };
                  });
                  """

        frida.load_script(js_code)
        frida.listen()

        frida.stop_frida_server()


    def main2():
        with Frida(with_use_init=False) as frida:
            # frida.spawn("com.example.networkapplication")
            frida.attach()

            # language=JavaScript
            js_code = """
                      Java.perform(function () {
                        let MainActivity = Java.use("com.example.networkapplication.MainActivity");
                        MainActivity["s"].implementation = function (message) {
                          console.log(`MainActivity.s is called: message=${message}`);
                          this["s"](message);
                        };
                      });
                      """

            frida.load_script(js_code)
            frida.listen()


    main1()
    main2()



from typing import Any


def list_dict_to_str(cookies_list_dict: list[dict[str, Any]]) -> str:
    return "; ".join([c["name"] + "=" + c["value"] for c in cookies_list_dict])


def str_to_dict(cookies_str: str) -> dict[str, str]:
    return {c.split("=")[0].strip(): c.split("=")[-1].strip() for c in cookies_str.split(";")}


def dict_to_str(cookies_dict: dict[str, str]) -> str:
    return "; ".join([k + "=" + v for k, v in cookies_dict.items()])



import re
from typing import Any

import bs4
import httpx
import parsel
import requests
import ujson

from xproject.xjavascript import execute_js_code_by_py_mini_racer


def jsonp_to_json(jsonp: str) -> dict[str, Any]:
    func_name = re.match(r"(?P<func_name>jQuery.*?)\(\{.*\}\)\S*", jsonp).groupdict()["func_name"]
    js_code = f"function {func_name}(o){{return o}};function sdk(){{return JSON.stringify({jsonp})}};"
    json_str = execute_js_code_by_py_mini_racer(js_code, func_name="sdk")
    json_obj = ujson.loads(json_str)
    return json_obj


def response_to_text(response: httpx.Response | requests.Response) -> Any:
    return response.text


def response_to_xpath(response: httpx.Response | requests.Response) -> parsel.Selector:
    return parsel.Selector(response.text)


def response_to_json(response: httpx.Response | requests.Response) -> Any:
    jsn = ujson.loads(response.text)
    return jsn


def response_to_soup(response: httpx.Response | requests.Response) -> bs4.BeautifulSoup:
    soup = bs4.BeautifulSoup(response.text, "html.parser")
    return soup



import os
from pathlib import Path
from tempfile import NamedTemporaryFile

import pillow_avif  # type: ignore
from PIL import Image


def image_to_jpg(
        image_file_path: str,
        /,
        jpg_file_path: str | None = None,
        quality: int = 100,
        keep_original: bool = False
) -> str | None:
    try:
        if jpg_file_path is None:
            image_file_prefix, _ = os.path.splitext(image_file_path)
            jpg_file_path = image_file_prefix + ".jpg"

        if image_file_path == jpg_file_path:
            return jpg_file_path

        jpg_dir_path = os.path.dirname(jpg_file_path)
        if not os.path.exists(jpg_dir_path):
            os.makedirs(jpg_dir_path, exist_ok=True)

        with Image.open(image_file_path) as image:
            if image.mode in ("RGBA", "LA"):
                background = Image.new("RGB", image.size, (255, 255, 255))
                background.paste(image, mask=image.split()[-1])
                image = background
            elif image.mode == "P":
                image.seek(0)  # image.n_frames
                image = image.convert("RGB")
            elif image.mode == "RGB":
                pass
            else:
                return None

            with NamedTemporaryFile(suffix=".jpg", delete=False, dir=jpg_dir_path) as ntf:
                temp_file_path = ntf.name
                image.save(temp_file_path, "JPEG", quality=quality)
            os.replace(temp_file_path, jpg_file_path)

        if not keep_original:
            p = Path(image_file_path)
            if p.exists() and str(p.resolve()) == str(p.absolute()):
                os.remove(image_file_path)

        return jpg_file_path
    except Exception as e:  # noqa
        return None


if __name__ == '__main__':
    from xproject.xurl import url_to_file_path

    input_file_path = url_to_file_path(
        # "https://img.jslink.com/FILE41f4c4a5db4e4007b09ae2dd2e08d853.JPG"
        # "https://www.cnncmall.com/photos/std-commodity/202505/802ae6c95ce01f49f1f9dada51169fa3.gif"
        # "https://www.cnncmall.com/photos/std-commodity/202506/324119dde7a483b0c54fdf7b2a5cbc2d.JPG"
        # "https://www.cnncmall.com/photos/std-commodity/202506/d09b671afd9ef8d1ef7abc86acf9d021.png"
        "https://fsyuncai.oss-cn-beijing.aliyuncs.com/2025-08-20/FILE6042f63720014c76bb4473e28710ad63.jpg"
    )
    output_file_path = image_to_jpg(input_file_path, keep_original=False)
    print(f"{input_file_path} => {output_file_path}")



from xproject.xadb import ADB
from xproject.xaliyun_oss import AliyunOSS
from xproject.xdataclasses.xaccount_dataclass import AccountDataclass
from xproject.xdbs.xmongo_db import MongoDB
from xproject.xdbs.xmysql_db import MysqlDB
from xproject.xdbs.xredis_db import RedisDB
from xproject.xfrida import Frida
from xproject.xhandler import Handler
from xproject.xlogger import get_logger, logger
from xproject.xspider.xenums.xdata_status_enum import DataStatusEnum
from xproject.xspider.xitems.xfield import Field
from xproject.xspider.xitems.xitem import Item
from xproject.xspider.xmodels.xmodel import Model
from xproject.xspider.xmodels.xsqlalchemy_model import SqlalchemyModel
from xproject.xtask import Task
from xproject.xwechat_ocr import WechatOCR

from . import scripts
from . import xadb
from . import xaliyun_oss
from . import xasyncio_priority_queue
from . import xbase64
from . import xbytes
from . import xcall
from . import xcommand
from . import xcookies
from . import xdata
from . import xdataclasses
from . import xdatetime
from . import xdbs
from . import xexceptions
from . import xfile
from . import xfrida
from . import xhandler
from . import xheaders
from . import xhtml
from . import xhttp
from . import ximage
from . import ximporter
from . import xjavascript
from . import xjson
from . import xlist
from . import xlogger
from . import xmath
from . import xmixins
from . import xnetwork
from . import xnotifier
from . import xpandas
from . import xrender
from . import xreverse
from . import xsignal
from . import xspider
from . import xsql
from . import xstring
from . import xtask
from . import xtypes
from . import xurl
from . import xvalidators
from . import xwechat_ocr

__all__ = [
    "ADB",
    "AccountDataclass",
    "AliyunOSS",
    "DataStatusEnum",
    "Field",
    "Frida",
    "Handler",
    "Item",
    "Model",
    "MongoDB",
    "MysqlDB",
    "RedisDB",
    "SqlalchemyModel",
    "Task",
    "WechatOCR",
    "get_logger",
    "logger",

    "scripts",
    "xadb",
    "xaliyun_oss",
    "xasyncio_priority_queue",
    "xbase64",
    "xbytes",
    "xcall",
    "xcommand",
    "xcookies",
    "xdata",
    "xdataclasses",
    "xdatetime",
    "xdbs",
    "xexceptions",
    "xfile",
    "xfrida",
    "xhandler",
    "xheaders",
    "xhtml",
    "xhttp",
    "ximage",
    "ximporter",
    "xjavascript",
    "xjson",
    "xlist",
    "xlogger",
    "xmath",
    "xmixins",
    "xnetwork",
    "xnotifier",
    "xpandas",
    "xrender",
    "xreverse",
    "xsignal",
    "xspider",
    "xsql",
    "xstring",
    "xtask",
    "xtypes",
    "xurl",
    "xvalidators",
    "xwechat_ocr",
]

import os

_assets_dir_name = "assets"  # noqa
ASSETS_DIR_PATH: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), _assets_dir_name)


def get_file_assets_dir_path(file_path: str, create_new_dir_path: bool = False) -> str:
    rel_path = os.path.relpath(os.path.abspath(file_path), start=ASSETS_DIR_PATH)
    new_dir_path = os.path.join(ASSETS_DIR_PATH, os.path.splitext(rel_path[3:])[0])  # 3 => ..\
    if create_new_dir_path:
        os.makedirs(new_dir_path, exist_ok=True)
    return new_dir_path


__all__ += ["get_file_assets_dir_path"]  # noqa



from abc import ABC

from xproject.xlogger import get_logger
from xproject.xmixins.xdir_path_mixin import DirPathMixin
from xproject.xmixins.xintrospection_mixin import IntrospectionMixin
from xproject.xmixins.xobject_mixins.xcontext_manager_object_mixin import ContextManagerObjectMixin
from xproject.xmixins.xrun_mixin import RunMixin


class Task(ContextManagerObjectMixin, RunMixin, DirPathMixin, IntrospectionMixin, ABC):
    logger = get_logger()



from typing import Any

import pandas as pd
from pymysql import connect
from pymysql.connections import Connection
from pymysql.cursors import Cursor

from xproject.xdbs.xdb import DB


class MysqlDB(DB):

    def __init__(
            self,
            *args: Any,
            host: str = "localhost",
            port: int = 3306,
            username: str | None = None,
            password: str | None = None,
            dbname: str,
            charset: str = "utf8mb4",
            **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._dbname = dbname
        self._charset = charset

        self._connection: Connection | None = None
        self._cursor: Cursor | None = None

    def open_connect(self) -> tuple[Connection, Cursor]:
        connection = connect(
            user=self._username,
            password=self._password,
            host=self._host,
            database=self._dbname,
            port=self._port,
            charset=self._charset,
        )
        cursor = connection.cursor()
        return connection, cursor

    @staticmethod
    def close_connect(connection: Connection | None = None, cursor: Cursor | None = None) -> None:
        if cursor is not None:
            cursor.close()

        if connection is not None:
            connection.close()

    def open(self) -> None:
        self._connection, self._cursor = self.open_connect()

    def close(self) -> None:
        self.close_connect(self._connection, self._cursor)

    @property
    def connection(self) -> Connection:
        return self._connection

    @property
    def cursor(self) -> Cursor:
        return self._cursor

    def query(
            self,
            sql: str,
            connection_cursor: tuple[Connection, Cursor] | None = None,
            use_new_connect: bool = True,  # Using False is in single thread
            return_df: bool = False
    ) -> list[dict[str, Any]] | pd.DataFrame:
        if connection_cursor is not None:
            connection, cursor = connection_cursor
        else:
            if use_new_connect:
                connection_cursor = self.open_connect()
                connection, cursor = connection_cursor
            else:
                connection, cursor = self._connection, self._cursor

        cursor.execute(sql)
        result = cursor.fetchall()
        columns = [d[0] for d in cursor.description]
        rows = [dict(zip(columns, r)) for r in result]

        if connection_cursor is not None:
            self.close_connect(connection, cursor)

        if return_df:
            return pd.DataFrame(rows)

        return rows



from typing import Any, Self

from xproject.xmixins.xobject_mixins.xcontext_manager_object_mixin import ContextManagerObjectMixin


class DB(ContextManagerObjectMixin):
    @classmethod
    def from_uri(cls, *args: Any, **kwargs: Any) -> Self:
        ins = super().create_instance(*args, **kwargs)
        return ins



from typing import Any, Self

from xproject.xcall import async_call_method
from xproject.xmixins.xobject_mixins.xobject_mixin import ObjectMixin


class AsyncCreateInstanceObjectMixin(ObjectMixin):
    @classmethod
    async def create_instance(cls, *args: Any, **kwargs: Any) -> Self:
        return super().create_instance(*args, **kwargs)

    async def _open(self) -> None:
        if self.is_closed:
            await async_call_method(method=self.open)
        self._is_closed = False

    async def _close(self) -> None:
        if not self.is_closed:
            await async_call_method(method=self.close)
        self._is_closed = True

    async def open(self) -> None:
        pass

    async def close(self) -> None:
        pass



from . import xdir_path_mixin
from . import xintrospection_mixin
from . import xmutable_mapping_mixin
from . import xobject_mixins
from . import xrun_mixin

__all__ = [
    "xdir_path_mixin",
    "xintrospection_mixin",
    "xmutable_mapping_mixin",
    "xobject_mixins",
    "xrun_mixin",
]



from typing import Any, Self
from urllib.parse import quote_plus

from pymongo import MongoClient
from pymongo.database import Database

from xproject.xdbs.xdb import DB


class MongoDB(DB):
    def __init__(
            self,
            *args: Any,
            host: str = "localhost",
            port: int = 27017,
            username: str | None = None,
            password: str | None = None,
            uri: str | None = None,
            dbname: str,
            **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._uri = uri
        self._dbname = dbname

        self._client: MongoClient | None = None
        self._db: Database | None = None

    @classmethod
    def from_uri(cls, uri: str, dbname: str) -> Self:
        ins = super().from_uri(**dict(uri=uri, dbname=dbname))
        return ins

    def open(self) -> None:
        if self._uri is not None:
            uri = self._uri
        else:
            if self._username is not None and self._password is not None:
                uri = "mongodb://%s:%s@%s:%s" % (
                    quote_plus(self._username), quote_plus(self._password), self._host, self._port
                )
            else:
                uri = "mongodb://%s:%s" % (self._host, self._port)

        self._client = MongoClient(uri)
        self._db = self._client[self._dbname]

    def close(self) -> None:
        self._db = None

        self._client.close()
        self._client = None

    @property
    def client(self) -> MongoClient:
        return self._client

    @property
    def db(self) -> Database:
        return self._db



import os
import re
import sys
from collections.abc import Buffer
from typing import Any, Final, cast
from urllib import parse

import furl
import httpx
import tldextract
from w3lib.url import canonicalize_url

from xproject.xheaders import get_default_headers


def get_furl_obj(url: str) -> furl.furl:
    return furl.furl(url)


def get_origin_path(url: str) -> str:
    """
    >>> get_origin_path("https://passport.jd.com/new/login.aspx?ReturnUrl=https%3A%2F%2Fwww.jd.com%2F")
    'https://passport.jd.com/new/login.aspx'

    :param url:
    :return:
    """
    furl_obj = get_furl_obj(url)
    origin_path = str(furl_obj.origin) + str(furl_obj.path)
    return origin_path


def get_parse_result(url: str) -> parse.ParseResult:
    parse_result = parse.urlparse(url)
    return parse_result


def is_valid(url: str) -> bool:
    """
    >>> is_valid("https://www.baidu.com/")
    True

    :param url:
    :return:
    """
    try:
        parse_result = get_parse_result(url)
        scheme, netloc = parse_result.scheme, parse_result.netloc
        if not scheme:
            return False
        if not netloc:
            return False
        if scheme not in ("http", "https"):
            return False
        return True
    except ValueError:
        return False


def quote(url: str, safe: str = "%;/?:@&=+$,", encoding: str = "utf-8") -> str:
    """
    >>> quote("https://www.baidu.com/s?wd=你好")
    'https://www.baidu.com/s?wd=%E4%BD%A0%E5%A5%BD'

    :param url:
    :param safe:
    :param encoding:
    :return:
    """
    return parse.quote_plus(url, safe=safe, encoding=encoding)


def unquote(url: str, encoding: str = "utf-8") -> str:
    """
    >>> unquote("https://www.baidu.com/s?wd=%E4%BD%A0%E5%A5%BD")
    'https://www.baidu.com/s?wd=你好'

    :param url:
    :param encoding:
    :return:
    """
    return parse.unquote_plus(url, encoding=encoding)


def encode(params: dict[str, Any]) -> str:
    """
    >>> encode({"a": "1", "b": "2"})
    'a=1&b=2'

    :param params:
    :return:
    """
    return parse.urlencode(params)


def decode(url: str) -> dict[str, str]:
    """
    >>> decode("xxx?a=1&b=2")
    {'a': '1', 'b': '2'}

    :param url:
    :return:
    """
    params = dict()

    lst = url.split("?", maxsplit=1)[-1].split("&")
    for i in lst:
        key, value = i.split("=", maxsplit=1)
        params[key] = unquote(value)

    return params


def join_url(base_url: str, url: str) -> str:
    """
    >>> join_url("https://www.baidu.com/", "/s?ie=UTF-8&wd=xproject")
    'https://www.baidu.com/s?ie=UTF-8&wd=xproject'

    :param base_url:
    :param url:
    :return:
    """
    return parse.urljoin(base_url, url)


def join_params(url: str, params: dict[str, Any]) -> str:
    """
    >>> join_params("https://www.baidu.com/s", {"wd": "你好"})
    'https://www.baidu.com/s?wd=%E4%BD%A0%E5%A5%BD'

    :param url:
    :param params:
    :return:
    """
    if not params:
        return url

    params = encode(params)
    separator = "?" if "?" not in url else "&"
    return url + separator + params


def get_params(url: str) -> dict[str, str]:
    """
    >>> get_params("https://www.baidu.com/s?wd=xproject")
    {'wd': 'xproject'}

    :param url:
    :return:
    """
    furl_obj = get_furl_obj(url)
    params = dict(furl_obj.query.params)
    return params


def get_param(url: str, key: str, default: Any | None = None) -> Any:
    """
    >>> get_param("https://www.baidu.com/s?wd=xproject", "wd")
    'xproject'

    :param url:
    :param key:
    :param default:
    :return:
    """
    params = get_params(url)
    param = params.get(key, default)
    return param


def get_url_params(url: str) -> tuple[str, dict[str, str]]:
    """
    >>> get_url_params("https://www.baidu.com/s?wd=xproject")
    ('https://www.baidu.com/s', {'wd': 'xproject'})

    :param url:
    :return:
    """
    root_url = ""
    params = dict()

    if "?" in url:
        root_url = url.split("?", maxsplit=1)[0]
        params = get_params(url)
    else:
        if re.search("[&=]", url) and not re.search("/", url):
            params = get_params(url)
        else:
            root_url = url

    return root_url, params


def get_domain(url: str) -> str:
    """
    >>> get_domain("https://www.baidu.com/s?wd=xproject")
    'baidu'

    :param url:
    :return:
    """
    er = tldextract.extract(url)
    domain = er.domain
    return domain


def get_subdomain(url: str) -> str:
    """
    >>> get_subdomain("https://passport.jd.com/new/login.aspx?ReturnUrl=https%3A%2F%2Fwww.jd.com%2F")
    'passport'

    :param url:
    :return:
    """
    er = tldextract.extract(url)
    subdomain = er.subdomain
    return subdomain


def canonicalize(url: str) -> str:
    """
    >>> canonicalize("https://www.baidu.com/s?wd=xproject")
    'https://www.baidu.com/s?wd=xproject'

    :param url:
    :return:
    """
    return canonicalize_url(url)


def url_to_file_path(
        url: str,
        headers: dict[str, Any] | None = None,
        file_path: str | None = None,
        dir_path: str | None = None,
        file_name: str | None = None,
        file_prefix: str | None = None,
        file_suffix: str | None = None,
        use_cache: bool = True,
        chunk_size: int = 64 * 1024
) -> str | None:
    if not is_valid(url):
        return None

    if headers is None:
        headers = get_default_headers()

    if file_path is None:
        if dir_path is None:
            sys_argv0 = sys.argv[0]
            if not os.path.isfile(sys_argv0):
                return None
            dir_path = os.path.dirname(sys_argv0)
        if file_name is None:
            if file_prefix is not None and file_suffix is not None:
                file_name = file_prefix + file_suffix
            else:
                if file_prefix is None:
                    file_prefix, _ = os.path.splitext(get_furl_obj(url).path.segments[-1])
                if file_suffix is None:
                    _, file_suffix = os.path.splitext(get_furl_obj(url).path.segments[-1])
                    if file_suffix is None or file_suffix == "":
                        response = httpx.head(url, headers=headers)
                        if (content_type := response.headers.get("content-type")) is not None:
                            # todo：Add more content_type.
                            content_type_to_file_suffix: Final[dict[str, str]] = {
                                "image/png": "png",
                                "image/gif": "gif",
                                "text/html;charset=utf-8": "html",
                                "text/javascript; charset=utf-8": "js",
                            }
                            file_ext = content_type_to_file_suffix[content_type]
                            file_suffix = "." + file_ext
                file_name = file_prefix + file_suffix
        file_path = os.path.join(dir_path, file_name)

    file_path = os.path.abspath(file_path)

    if use_cache:
        if os.path.exists(file_path):
            return file_path

    dir_path = os.path.dirname(file_path)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    try:
        with httpx.Client(timeout=None, follow_redirects=True) as client:
            with client.stream("GET", url, headers=headers) as response:
                response.raise_for_status()
                with open(file_path, "wb") as f:
                    for chunk in response.iter_bytes(chunk_size=chunk_size):
                        f.write(chunk)
    except Exception as e:  # noqa
        file_path = None

    return file_path



import hashlib
from typing import Any, Literal

ALGO_TYPE = Literal[
    "md5", "sha1", "sha224", "sha256", "sha384", "sha512", "blake2b", "blake2s", "sha3_224", "sha3_256", "sha3_384",
    "sha3_512", "shake_128", "shake_256"
]


def gen_data_id(
        *args: Any,
        keys: list | None = None, item: dict | None = None,
        algo_type: ALGO_TYPE = "sha256"
) -> str:
    """
    >>> gen_data_id("123456")
    '8d969eef6ecad3c29a3a629280e686cf0c3f5d5a86aff3ca12020c923adc6c92'

    :param args:
    :param keys:
    :param item:
    :param algo_type:
    :return:
    """
    m = hashlib.new(algo_type)
    if args:
        values = args
    elif keys is not None and item is not None:
        if isinstance(keys, list) and isinstance(item, dict):
            values = [item[k] for k in keys if k in item]
        else:
            raise TypeError(
                f"Invalid type for 'keys': "
                f"Expected `list | None`, "
                f"but got {type(keys).__name__} (value: {keys!r})\n"
                f"Invalid type for 'item': "
                f"Expected `dict | None`, "
                f"but got {type(item).__name__} (value: {item!r})"
            )
    elif item is not None:
        values = [item[k] for k in sorted(item.keys())]
    else:
        raise ValueError(
            f"Either args: {args!r} or keys: {keys!r} and item: {item!r} or item: {item!r} must be provided"
        )

    data = list(map(lambda x: str(x), values))

    for i in data:
        m.update(i.encode())

    data_id = m.hexdigest()
    return data_id



import html
from typing import Any

from htmlmin import minify


def compress_html(*args: Any, **kwargs: Any) -> str:
    return minify(*args, **kwargs)


def escape(text: str) -> str:
    """
    >>> escape("<")
    '&lt;'

    """
    return html.escape(text)


def unescape(text: str) -> str:
    """
    >>> unescape("&lt;")
    '<'

    """
    return html.unescape(text)



import collections.abc as c
import typing as t
from inspect import iscoroutinefunction, signature

import blinker


class Signal(blinker.Signal):
    @staticmethod
    def _get_filter_kwargs(sender, receiver, **kwargs):
        kwargs.setdefault("sender", sender)
        sig = signature(receiver)
        return {k: v for k, v in kwargs.items() if k in sig.parameters}

    def send(
            self,
            sender: t.Any | None = None,
            /,
            *,
            _async_wrapper: c.Callable[
                                [c.Callable[..., c.Coroutine[t.Any, t.Any, t.Any]]], c.Callable[..., t.Any]
                            ]
                            | None = None,
            **kwargs: t.Any,
    ) -> list[tuple[c.Callable[..., t.Any], t.Any]]:
        """Call all receivers that are connected to the given ``sender``
        or :data:`ANY`. Each receiver is called with ``sender`` as a positional
        argument along with any extra keyword arguments. Return a list of
        ``(receiver, return value)`` tuples.

        The order receivers are called is undefined, but can be influenced by
        setting :attr:`set_class`.

        If a receiver raises an exception, that exception will propagate up.
        This makes debugging straightforward, with an assumption that correctly
        implemented receivers will not raise.

        :param sender: Call receivers connected to this sender, in addition to
            those connected to :data:`ANY`.
        :param _async_wrapper: Will be called on any receivers that are async
            coroutines to turn them into sync callables. For example, could run
            the receiver with an event loop.
        :param kwargs: Extra keyword arguments to pass to each receiver.

        .. versionchanged:: 1.7
            Added the ``_async_wrapper`` argument.
        """
        if self.is_muted:
            return []

        results = []

        for receiver in self.receivers_for(sender):
            filter_kwargs = self._get_filter_kwargs(sender, receiver, **kwargs)

            if iscoroutinefunction(receiver):
                if _async_wrapper is None:
                    raise RuntimeError("Cannot send to a coroutine function.")

                result = _async_wrapper(receiver)(**filter_kwargs)
            else:
                result = receiver(**filter_kwargs)

            results.append((receiver, result))

        return results

    async def send_async(
            self,
            sender: t.Any | None = None,
            /,
            *,
            _sync_wrapper: c.Callable[
                               [c.Callable[..., t.Any]], c.Callable[..., c.Coroutine[t.Any, t.Any, t.Any]]
                           ]
                           | None = None,
            **kwargs: t.Any,
    ) -> list[tuple[c.Callable[..., t.Any], t.Any]]:
        """Await all receivers that are connected to the given ``sender``
        or :data:`ANY`. Each receiver is called with ``sender`` as a positional
        argument along with any extra keyword arguments. Return a list of
        ``(receiver, return value)`` tuples.

        The order receivers are called is undefined, but can be influenced by
        setting :attr:`set_class`.

        If a receiver raises an exception, that exception will propagate up.
        This makes debugging straightforward, with an assumption that correctly
        implemented receivers will not raise.

        :param sender: Call receivers connected to this sender, in addition to
            those connected to :data:`ANY`.
        :param _sync_wrapper: Will be called on any receivers that are sync
            callables to turn them into async coroutines. For example,
            could call the receiver in a thread.
        :param kwargs: Extra keyword arguments to pass to each receiver.

        .. versionadded:: 1.7
        """
        if self.is_muted:
            return []

        results = []

        for receiver in self.receivers_for(sender):
            filter_kwargs = self._get_filter_kwargs(sender, receiver, **kwargs)

            if not iscoroutinefunction(receiver):
                if _sync_wrapper is None:
                    raise RuntimeError("Cannot send to a non-coroutine function.")

                result = await _sync_wrapper(receiver)(**filter_kwargs)
            else:
                result = await receiver(**filter_kwargs)

            results.append((receiver, result))

        return results


__all__ = [
    "Signal",
]



import os
import re
import sys
from typing import Any

from rich.console import Console
from rich.syntax import Syntax

from xproject.xlist import flatten_list

console = Console()


class GenerateInitPy:
    @staticmethod
    def _generate_init_py_by_dir_path(dir_path: str) -> int:
        dir_path = os.path.abspath(dir_path)
        if not os.path.isdir(dir_path):
            return 1

        if os.path.basename(dir_path) == "__pycache__":
            return 0

        module_names = []

        init_py_file_content_addition_a: str | None = None
        init_py_file_content_addition_b: str | None = None
        init_py_file_content_addition_c: str | None = None

        for file_name in sorted(os.listdir(dir_path)):
            file_path = os.path.join(dir_path, file_name)
            if os.path.isfile(file_path):
                if file_name == "xproject.__init__.py":
                    try:
                        with open(file_path, "r", encoding="utf-8") as file:
                            content = file.read()

                            init_py_file_content_addition_a = "\n".join(
                                sorted(
                                    re.findall(r"(from .*? import .*?)\n", content, re.DOTALL)
                                )
                            )
                            init_py_file_content_addition_b = ",\n    ".join([
                                f"\"{module_name}\"" for module_name in
                                sorted(
                                    flatten_list([i.split(", ") if ", " in i else i for i in
                                                  re.findall(r"from .*? import (.*?)\n", content, re.DOTALL)])
                                )
                            ])
                            init_py_file_content_addition_c = re.findall(
                                r"# script\n(.*?)\n# script\n", content, re.DOTALL
                            )[0]

                    except Exception:  # noqa
                        pass
                else:
                    if file_name.endswith(".py") and file_name not in ("__init__.py", "__main__.py"):
                        module_names.append(os.path.splitext(file_name)[0])
            elif os.path.isdir(file_path):
                if file_name != "__pycache__" and any(map(lambda x: x.endswith(".py"), os.listdir(file_path))):
                    module_names.append(file_name)

        if not module_names:
            return 0

        init_py_file_content_a = "\n".join(f"from . import {module_name}" for module_name in module_names)
        init_py_file_content_b = ",\n    ".join(f"\"{module_name}\"" for module_name in module_names)

        if init_py_file_content_addition_a is not None and init_py_file_content_addition_b is not None and init_py_file_content_addition_c is not None:
            init_py_file_content = f"""{init_py_file_content_addition_a}

{init_py_file_content_a}

__all__ = [
    {init_py_file_content_addition_b},

    {init_py_file_content_b},
]

{init_py_file_content_addition_c}
"""
        else:
            init_py_file_content = f"""{init_py_file_content_a}

__all__ = [
    {init_py_file_content_b},
]
"""

        init_py_file_path = os.path.join(dir_path, "__init__.py")
        console.print(f"[bold green]{init_py_file_path}[/bold green]")

        with open(init_py_file_path, "w", encoding="utf-8") as f:
            f.write(init_py_file_content)
            syntax = Syntax(init_py_file_content, "python", theme="monokai", line_numbers=True)
            console.print(syntax)

        return 0

    @classmethod
    def _generate_init_py_by_root_dir_path(cls, root_dir_path: str) -> int:
        root_dir_path = os.path.abspath(root_dir_path)
        if not os.path.isdir(root_dir_path):
            return 1

        if os.path.basename(root_dir_path) == "__pycache__":
            return 0

        for root, dirs, files in os.walk(root_dir_path):
            if cls._generate_init_py_by_dir_path(root) != 0:
                return 1

        return 0

    @classmethod
    def main(cls, *args: Any) -> None:
        dir_path = os.path.abspath(args[0] if len(args) >= 1 else sys.argv[1] if len(sys.argv) > 1 else ".")
        sys.exit(cls._generate_init_py_by_root_dir_path(dir_path))



from abc import ABC
from atexit import register
from types import TracebackType
from typing import Any, Self

from xproject.xcall import call_method
from xproject.xmixins.xobject_mixins.xcreate_instance_object_mixin import CreateInstanceObjectMixin


class ContextManagerObjectMixin(CreateInstanceObjectMixin, ABC):
    @classmethod
    def create_instance(cls, *args: Any, auto_call: bool = True, **kwargs: Any) -> Self:
        instance = super().create_instance(*args, **kwargs)
        if auto_call:
            call_method(method=instance._open)
            register(lambda: call_method(method=instance._close))
        return instance

    def __enter__(self) -> Self:
        call_method(method=self._open)
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None
    ) -> None:
        call_method(method=self._close)



import os
from abc import ABCMeta
from inspect import getmodule
from typing import Any, Self


class DirPathMixinMeta(ABCMeta):
    def __new__(
            mcls: type[Self],
            name: str,
            bases: tuple[type, ...],
            namespace: dict[str, Any],
            /,
            **kwargs: Any
    ) -> Self:
        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        if namespace.get("dir_path") is None:
            if (module := getmodule(cls)) is not None and hasattr(module, "__file__"):
                cls.dir_path = os.path.dirname(os.path.abspath(module.__file__))
            else:
                cls.dir_path = None

        return cls


class DirPathMixin(metaclass=DirPathMixinMeta):
    dir_path: str | None



import dataclasses
import re
import subprocess

from charset_normalizer import from_bytes

from xproject.xlogger import Logger


@dataclasses.dataclass(frozen=True)
class SubprocessPopenResult:
    returncode: int  # noqa
    content: bytes
    text: str


def execute_cmd_code_by_subprocess_popen(
        cmd_code: str,
        encoding: str | None = None,
        logger: Logger | None = None
) -> SubprocessPopenResult:
    if logger:
        logger.debug(f'''$ {cmd_code}''')

    process = subprocess.Popen(
        cmd_code, shell=True,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    content = bytes()
    text = str()
    for line in process.stdout:
        content += line
        txt = "> "
        if encoding:
            txt += line.decode(encoding)
        elif best_match := from_bytes(line).best():
            txt += line.decode(best_match.encoding)
        else:
            txt += str(line)
        txt = re.sub(r"\r\n", "\n", txt, re.DOTALL)
        txt = re.sub(r"\n$", "", txt, re.DOTALL)
        if logger:
            logger.debug(txt)
        text += re.sub(r"^> ", "", txt, re.DOTALL) + "\n"

    returncode = process.wait()

    return SubprocessPopenResult(returncode, content, text)


@dataclasses.dataclass(frozen=True)
class SubprocessRunResult:
    result: subprocess.CompletedProcess
    content: bytes
    text: str


def execute_cmd_code_by_subprocess_run(
        cmd_code: str,
        encoding: str | None = None,
        logger: Logger | None = None
) -> SubprocessRunResult:
    if logger:
        logger.debug(f'''$ {cmd_code}''')

    result = subprocess.run(cmd_code, shell=True, capture_output=True)

    content = bytes()
    text = "> "
    for i in [result.stdout, result.stderr]:
        if i:
            content += i
            if encoding:
                text += i.decode(encoding)
            elif best_match := from_bytes(i).best():
                text += i.decode(best_match.encoding)
            else:
                text += str(i)
            text += "\n"
    text = re.sub(r"\r\n", "\n", text, re.DOTALL)
    text = re.sub(r"\n$", "", text, re.DOTALL)
    if logger:
        logger.debug(text)
    text = re.sub(r"^> ", "", text, re.DOTALL)

    return SubprocessRunResult(result, content, text)


if __name__ == '__main__':
    from xproject.xlogger import get_logger

    print(execute_cmd_code_by_subprocess_popen("ping www.baidu.com", "cp936", get_logger()).text)
    print(execute_cmd_code_by_subprocess_run("ping www.baidu.com", "cp936", get_logger()).text)
    print(execute_cmd_code_by_subprocess_popen("pip show frida", logger=get_logger()).text)



from typing import Any

import sqlparse


def format_sql(sql: str, **kwargs: Any) -> str:
    if not kwargs:
        kwargs = dict(
            reindent=True,
            keyword_case="upper",
            identifier_case="lower",
            strip_comments=True
        )
    sql = sqlparse.format(sql, **kwargs)
    return sql



from collections.abc import Callable, Awaitable
from inspect import iscoroutinefunction, ismethod
from typing import Any


async def async_call_method(
        *,
        method: Callable[..., Awaitable[Any] | Any] | None = None,
        obj: Any = None,
        method_name: str | None = None,
        method_args: tuple[Any, ...] | None = None,
        method_kwargs: dict[str, Any] | None = None
) -> Any:
    if method is None:
        if obj is not None and method_name is not None:
            method = getattr(obj, method_name, None)
    if method is None:
        return None
    if method_args is None:
        method_args = tuple()
    if method_kwargs is None:
        method_kwargs = dict()
    if iscoroutinefunction(method):
        return await method(*method_args, **method_kwargs)
    if ismethod(method):
        return method(*method_args, **method_kwargs)
    raise TypeError(
        f"Invalid type for 'method': "
        f"Expected `Callable[..., Awaitable[Any] | Any] | None`, "
        f"but got {type(method).__name__} (value: {method!r})"
    )


def call_method(
        *,
        method: Callable[..., Any] | None = None,
        obj: Any = None,
        method_name: str | None = None,
        method_args: tuple[Any, ...] | None = None,
        method_kwargs: dict[str, Any] | None = None
) -> Any:
    if method is None:
        if obj is not None and method_name is not None:
            method = getattr(obj, method_name, None)
    if method is None:
        return None
    if method_args is None:
        method_args = tuple()
    if method_kwargs is None:
        method_kwargs = dict()
    if not iscoroutinefunction(method) and (ismethod(method) or callable(method)):
        return method(*method_args, **method_kwargs)
    raise TypeError(
        f"Invalid type for 'method': "
        f"Expected `Callable[..., Any] | None`, "
        f"but got {type(method).__name__} (value: {method!r})"
    )



import os
import platform
import zipfile
from typing import Literal, cast


def compress_dir_path(
        dir_path: str,
        compress_type: Literal["zip"] = "zip"
) -> str | None:
    dir_path = os.path.abspath(dir_path)

    if not os.path.isdir(dir_path):
        return None

    valid_compress_type = ["zip"]
    if compress_type not in valid_compress_type:
        return None

    compress_file_path = os.path.join(
        os.path.dirname(dir_path), os.path.basename(dir_path) + "." + compress_type
    )

    if compress_type == "zip":
        with zipfile.ZipFile(compress_file_path, "w", zipfile.ZIP_DEFLATED) as zf:
            for root, dirs, files in os.walk(dir_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    rel_file_path = os.path.relpath(file_path, dir_path)
                    zf.write(file_path, arcname=rel_file_path)

        return compress_file_path

    return None


def get_file_paths_and_dir_paths(path: str) -> tuple[list[str], list[str]]:
    file_paths = []
    dir_paths = []

    path = os.path.abspath(path)

    with os.scandir(path) as entries:
        for entry in entries:
            System = Literal["Windows", "Linux", "Darwin"]
            system: System = cast(System, platform.system())
            if system == "Windows":
                from nt import DirEntry
            elif system == "Linux":
                from posix import DirEntry
            elif system == "Darwin":
                from posix import DirEntry
            else:
                raise TypeError(
                    f"Invalid type for 'system': "
                    f"Expected `Literal[\"Windows\",\"Linux\",\"Darwin\"]`, "
                    f"but got {type(system).__name__!r} (value: {system!r})"
                )

            entry: DirEntry
            if entry.is_file():
                file_paths.append(entry.path)
            elif entry.is_dir():
                dir_paths.append(entry.path)
                sub_file_paths, sub_dir_paths = get_file_paths_and_dir_paths(entry.path)
                file_paths.extend(sub_file_paths)
                dir_paths.extend(sub_dir_paths)

    return file_paths, dir_paths


if __name__ == '__main__':
    print(compress_dir_path("."))
    print(get_file_paths_and_dir_paths("."))



from functools import partial
from typing import TYPE_CHECKING

import pandas as pd

from xproject.xcommand import SubprocessPopenResult, SubprocessRunResult, execute_cmd_code_by_subprocess_popen, \
    execute_cmd_code_by_subprocess_run
from xproject.xlogger import Logger, get_logger
from xproject.xnetwork import get_host


class ADB:
    def __init__(
            self,
            logger: Logger | None = None
    ) -> None:
        self.logger: Logger | None = None
        if logger is None:
            logger = get_logger()
        else:
            self.logger = logger
        self._logger = logger

    def open_log(self) -> None:
        if self.logger is None:
            self.logger = self._logger

    def close_log(self) -> None:
        if self.logger is not None:
            self.logger = None

    def execute_cmd_code_by_subprocess_popen(self, cmd_code: str, encoding: str | None = None) -> SubprocessPopenResult:
        if self.logger is None:
            return execute_cmd_code_by_subprocess_popen(cmd_code, encoding=encoding)
        return execute_cmd_code_by_subprocess_popen(cmd_code, encoding=encoding, logger=self.logger)

    def execute_cmd_code_by_subprocess_run(self, cmd_code: str, encoding: str | None = None) -> SubprocessRunResult:
        if self.logger is None:
            return execute_cmd_code_by_subprocess_run(cmd_code, encoding=encoding)
        return execute_cmd_code_by_subprocess_run(cmd_code, encoding=encoding, logger=self.logger)

    def get_android_version(self) -> str:
        return self.execute_cmd_code_by_subprocess_run("adb shell getprop ro.build.version.release").text

    def get_android_sdk(self) -> str:
        return self.execute_cmd_code_by_subprocess_run("adb shell getprop ro.build.version.sdk").text

    def get_cpu_abi(self) -> str:
        return self.execute_cmd_code_by_subprocess_run("adb shell getprop ro.product.cpu.abi").text

    def set_reverse_tcp(self, local_port: int = 8888, remote_port: int = 8888) -> None:
        self.execute_cmd_code_by_subprocess_run(f"adb reverse tcp:{local_port} tcp:{remote_port}")

    def get_reverse_tcp_list(self) -> list[str]:
        lst = self.execute_cmd_code_by_subprocess_run("adb reverse --list").text.split("\n")
        return [i for i in lst if i]

    def clear_reverse_tcp(self, local_port: int = 8888) -> None:
        self.execute_cmd_code_by_subprocess_run(f"adb reverse --remove tcp:{local_port}")

    def set_global_http_proxy(self, host: str, port: int = 8888) -> None:
        http_proxy = f"{host}:{port}"
        self.execute_cmd_code_by_subprocess_run(f"adb shell settings put global http_proxy {http_proxy}")

    def get_global_http_proxy(self) -> str:
        return self.execute_cmd_code_by_subprocess_run("adb shell settings get global http_proxy").text

    def clear_global_http_proxy(self) -> None:
        self.execute_cmd_code_by_subprocess_run("adb shell settings delete global http_proxy")
        self.execute_cmd_code_by_subprocess_run("adb shell settings delete global global_http_proxy_host")
        self.execute_cmd_code_by_subprocess_run("adb shell settings get global global_http_proxy_port")

    def reboot(self) -> None:
        self.execute_cmd_code_by_subprocess_run("adb reboot")

    def get_process_names_df(self, process_name: str | None = None) -> pd.DataFrame:
        lines = self.execute_cmd_code_by_subprocess_run("adb shell ps -A").text.splitlines()
        df = pd.DataFrame(
            [line.split(maxsplit=len(lines[0].split()) - 1) for line in lines[1:]], columns=lines[0].split()
        )
        if process_name:
            df = df[df["NAME"].str.contains(process_name)]
        return df

    if TYPE_CHECKING:
        @staticmethod
        def get_pc_local_host() -> str:
            ...
    else:
        get_pc_local_host = staticmethod(partial(get_host, host_type="local"))


if __name__ == '__main__':
    adb = ADB()

    adb.open_log()

    # print(adb.get_android_version())
    # print(adb.get_android_sdk())
    # print(adb.get_cpu_abi())

    # adb.set_reverse_tcp()
    # print(adb.get_reverse_tcp_list())
    # adb.clear_reverse_tcp()

    # adb.set_global_http_proxy("127.0.0.1", 8888)
    # print(adb.get_global_http_proxy())
    # adb.clear_global_http_proxy()

    adb.reboot()

    adb.close_log()

    # print(adb.get_process_names_df())
    # adb.get_process_names_df().to_excel("output.xlsx", index=False)



import json
from datetime import datetime
from decimal import Decimal
from typing import Any


class JSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(obj, Decimal):
            return float(obj)
        return super().default(obj)



from abc import ABC, abstractmethod
from typing import Any

from xproject.xlogger import Logger


class RunMixin(ABC):
    logger: Logger

    @abstractmethod
    def action(self, *args, **kwargs):
        pass

    def run_ok_method(self, result: Any) -> Any:
        pass

    def run_error_method(self, exception: Exception) -> bool | None:
        """

        :param exception:
        :return: None => raise exception
                 True => ignore exception and no logger
                 False => ignore exception and logger
        """
        pass

    def run(self, *args: Any, **kwargs: Any, ) -> Any:
        try:
            result = self.action(*args, **kwargs)
        except Exception as e:
            ret = self.run_error_method(e)
            if ret is None:
                raise e
            if ret is True:
                pass
            if ret is False:
                self.logger.exception(e)
        else:
            self.run_ok_method(result)
            return result
        return None

    def loop_run(self, *args: Any, **kwargs: Any) -> None:
        run_times = 0
        while True:
            run_times += 1
            self.logger.debug(f"loop_run run_times: {run_times} start")
            self.run(*args, **kwargs)
            self.logger.debug(f"loop_run run_times: {run_times} end")



from abc import ABC

from xproject.xlogger import get_logger
from xproject.xmixins.xobject_mixins.xcontext_manager_object_mixin import ContextManagerObjectMixin
from xproject.xmixins.xrun_mixin import RunMixin


class Handler(ContextManagerObjectMixin, RunMixin, ABC):
    logger = get_logger()



from http.cookies import SimpleCookie
from typing import Any

import scrapy
from fake_useragent import UserAgent


def get_default_headers(**kwargs: Any) -> dict[str, str]:
    if not kwargs:
        kwargs = dict(platforms=["desktop"])
    user_agent = UserAgent(**kwargs)

    headers = {
        "User-Agent": user_agent.random
    }
    return headers


def update_headers(
        headers: dict[str, str] | None = None,
        default_headers: dict[str, str] | None = None
) -> dict[str, str]:
    if headers is None:
        headers = {}
    if default_headers is None:
        default_headers = get_default_headers()

    for k, v in default_headers.items():
        if k not in headers:
            headers[k] = v

    return headers


def scrapy_response_headers_to_cookies(response: scrapy.http.Response) -> dict[str, str]:
    cookies = {}
    for header in response.headers.getlist("Set-Cookie"):
        c = SimpleCookie()
        c.load(header.decode())
        for key, morsel in c.items():
            cookies[key] = morsel.value
    return cookies


if __name__ == '__main__':
    print(get_default_headers())
    print(update_headers({"a": "1"}, {"b": "2"}))



from __future__ import annotations

import hashlib
import json
import re
from abc import ABCMeta
from collections.abc import Iterable
from contextlib import contextmanager
from typing import Any, Self, TYPE_CHECKING, cast

from sqlalchemy import Engine, Column, Integer, String, DateTime, SmallInteger, JSON, func
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.types import Text, TypeDecorator

from xproject.xjson import JSONEncoder
from xproject.xspider.xenums.xdata_status_enum import DataStatusEnum
from xproject.xspider.xmodels.xmodel import Model
from xproject.xstring import camel_to_snake

if TYPE_CHECKING:
    from xproject.xspider.xitems.xitem import Item

Base = declarative_base()


class MediumText(TypeDecorator):
    impl = Text

    def load_dialect_impl(self, dialect: Dialect) -> TypeEngine[Any]:
        if dialect.name == "mysql":
            # MEDIUMTEXT: Maximum length 16 MB (16,777,215 bytes)
            return dialect.type_descriptor(Text(length=16777215))
        return super().load_dialect_impl(dialect)


class LongText(TypeDecorator):
    impl = Text

    def load_dialect_impl(self, dialect: Dialect) -> TypeEngine[Any]:
        if dialect.name == "mysql":
            # LONGTEXT: Maximum length 4 GB (4,294,967,295 bytes)
            return dialect.type_descriptor(Text(length=4294967295))
        return super().load_dialect_impl(dialect)


class SqlalchemyModelMeta(type(Base), ABCMeta):
    def __new__(mcs, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> type:
        if attrs.get("__abstract__") is True:
            return super().__new__(mcs, name, bases, attrs)

        table_name = attrs.get("__tablename__")
        if table_name is None:
            table_name = camel_to_snake(name)
            attrs["__tablename__"] = table_name
        else:
            if not (isinstance(table_name, str) and table_name):
                raise ValueError(f"Illegal assignment {name}.__tablename__")

        engine = attrs.get("__engine__")
        if not isinstance(engine, Engine):
            raise ValueError(f"Illegal assignment {name}.__engine__")

        data_columns = attrs.get("__data_columns__")
        if not (isinstance(data_columns, list) and all(map(lambda x: isinstance(x, str), data_columns))):
            raise ValueError(f"Illegal assignment {name}.__data_columns__")

        session_factory = attrs.get("__session_factory__")
        if session_factory is None:
            session_factory = sessionmaker(bind=engine)
            attrs["__session_factory__"] = session_factory
        else:
            if not isinstance(session_factory, sessionmaker):
                raise ValueError(f"Illegal assignment {name}.__session_factory__")

        return super().__new__(mcs, name, bases, attrs)


class SqlalchemyModel(Base, Model, metaclass=SqlalchemyModelMeta):
    __abstract__ = True

    __tablename__: str
    __engine__: Engine
    __data_columns__: list[str]
    __session_factory__: sessionmaker

    id = Column(Integer, comment="id", primary_key=True, autoincrement=True)
    data_id = Column(String(64), comment="data id", nullable=False, unique=True)
    data_columns = Column(JSON, comment="data columns", nullable=False)
    data_status = Column(SmallInteger, comment="data status", default=DataStatusEnum.OK.value, index=True)
    data_create_time = Column(DateTime, comment="data create time", server_default=func.now())
    data_update_time = Column(DateTime, comment="data update time", server_default=func.now(), onupdate=func.now())

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} (id={self.id}, data_id={self.data_id})>"

    def to_row(self) -> dict[str, Any]:
        data = {column.name: getattr(self, column.name) for column in cast(Iterable, self.__table__.columns)}
        jsn = json.dumps(data, ensure_ascii=False, cls=JSONEncoder)
        row = json.loads(jsn)
        return row

    @classmethod
    @contextmanager
    def get_session(cls, read_only: bool = False):
        session = cls.__session_factory__()
        try:
            yield session
            if not read_only:
                session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    @classmethod
    def get_table_name(cls) -> str:
        return cls.__tablename__

    @classmethod
    def gen_data_id(cls, data: dict[str, Any] | Item) -> str:
        data_column_value_strs = []
        for data_column in cls.__data_columns__:
            if data_column in data:
                data_column_value_str = str(data[data_column])
                data_column_value_strs.append(data_column_value_str)
            else:
                raise ValueError(f"data must provide the '{data_column}' field")
        data_id = hashlib.sha256("".join(data_column_value_strs).encode()).hexdigest()
        return data_id

    @classmethod
    def get_data_id(cls, data: dict[str, Any] | Item) -> str:
        if "data_id" in data:
            if isinstance(data["data_id"], str) and re.match(r"^[0-9a-f]{64}$", data["data_id"]) is not None:
                return data["data_id"]
        return cls.gen_data_id(data)

    @classmethod
    def get_ins_by_data_id(cls, data: dict[str, Any] | Item) -> Self | None:
        with cls.get_session(read_only=True) as session:
            data_id = cls.get_data_id(data)
            ins: Self = session.query(cls).filter_by(data_id=data_id).one_or_none()
            if ins is None:
                return None
            session.expunge(ins)
            return ins

    @classmethod
    def get_row_by_data_id(cls, data: dict[str, Any] | Item) -> dict[str, Any] | None:
        with cls.get_session(read_only=True) as session:
            data_id = cls.get_data_id(data)
            ins: Self = session.query(cls).filter_by(data_id=data_id).one_or_none()
            if ins is None:
                return None
            return ins.to_row()

    @classmethod
    def save(cls, data: dict[str, Any] | Item) -> bool:
        with cls.get_session() as session:
            data_id = cls.get_data_id(data)
            ins: Self = session.query(cls).filter_by(data_id=data_id).one_or_none()
            insert = ins is None

            if insert:
                # insert
                ins = cls(
                    data_id=data_id,
                    data_columns=cls.__data_columns__,
                    **data
                )
                session.add(ins)
            else:
                # update
                for k, v in data.items():
                    setattr(ins, k, v)

            return insert

    @classmethod
    def create_table(cls) -> None:
        Base.metadata.create_all(cls.__engine__)

    @classmethod
    def columns(cls) -> list[str]:
        return [c.name for c in cast(Iterable, cls.__table__.columns)]



import re


def camel_to_snake(string: str) -> str:
    """
    >>> camel_to_snake('CamelCaseString')
    'camel_case_string'

    """
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    string = re.sub("([a-z0-9])([A-Z])", r"\1_\2", string).lower()
    return string


def snake_to_camel(string: str) -> str:
    """
    >>> snake_to_camel('snake_case_string')
    'SnakeCaseString'

    """
    return "".join(i.capitalize() for i in string.split("_"))


if __name__ == '__main__':
    print(camel_to_snake("CamelCaseString"))
    print(snake_to_camel("snake_case_string"))



from typing import Any


def split_list(data: list[Any], num_parts: int | None = None, part_size: int | None = None) -> list[Any]:
    if not data:
        return []

    if num_parts:
        avg_size = len(data) // num_parts
        remainder = len(data) % num_parts

        result = []
        result.extend([data[start:start + avg_size] for start in range(0, len(data) - avg_size - remainder, avg_size)])
        result.append(data[len(data) - avg_size - remainder:])
        return result

    elif part_size:
        return [data[i:i + part_size] for i in range(0, len(data), part_size)]

    else:
        raise ValueError(
            f"Either num_parts: {num_parts!r} or part_size: {part_size!r} must be provided"
        )


def flatten_list(data: list[Any]) -> list[Any]:
    return sum((flatten_list(x) if isinstance(x, list) else [x] for x in data), [])


if __name__ == '__main__':
    print(flatten_list([1, 2, [3, 4], [5, 6, 7]]))



from collections.abc import Buffer

from ujson import dumps


def to_bytes(
        data: bytes | str | dict,
        /, *,
        dict_sort_keys: bool = False,
        encoding: str = "utf-8"
) -> Buffer:
    if isinstance(data, bytes):
        return data
    if isinstance(data, str):
        return data.encode(encoding)
    if isinstance(data, dict):
        return dumps(data, sort_keys=dict_sort_keys).encode(encoding)
    raise TypeError(
        f"Invalid type for 'data': "
        f"Expected `bytes | str | dict`, "
        f"but got {type(data).__name__!r} (value: {data!r})"
    )



from inspect import currentframe


class IntrospectionMixin:
    def __repr__(self) -> str:
        return self.__class__.__name__

    __str__ = __repr__

    @property
    def method_name(self) -> str:
        frame = currentframe().f_back
        return frame.f_code.co_name

    @property
    def method_qualname(self) -> str:
        frame = currentframe().f_back
        method_name = frame.f_code.co_name
        method = getattr(self, method_name)
        return method.__qualname__



from asyncio import PriorityQueue, TimeoutError, wait_for
from typing import Any


class AsyncioPriorityQueue(PriorityQueue):
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)

    async def get(self) -> Any:
        coroutine = super().get()
        try:
            return await wait_for(coroutine, timeout=0.1)
        except TimeoutError:
            return None



import arrow


def now(fmt: str = "YYYY-MM-DD HH:mm:ss") -> str:
    return arrow.now().format(fmt)


def delta(start_datetime_str: str, end_datetime_str: str) -> int:
    start_datetime = arrow.get(start_datetime_str).datetime
    end_datetime = arrow.get(end_datetime_str).datetime
    seconds = (end_datetime - start_datetime).total_seconds()
    return int(seconds)


if __name__ == '__main__':
    print(now())
    print(delta("2025-08-01 00:00:00", "2025-08-01 11:11:11"))



import os
import sys
from typing import Any, cast

from xproject.xrender import Render
from xproject.xstring import snake_to_camel, camel_to_snake


class RenderTaskPy(Render):
    @classmethod
    def main(cls, *args: Any) -> None:
        render_file_path = os.path.abspath(
            args[0] if len(args) >= 1 else sys.argv[1] if len(sys.argv) > 1 else "task.py"
        )
        class_name = snake_to_camel(os.path.splitext(os.path.basename(render_file_path))[0])
        instance_name = camel_to_snake(class_name)
        render_data = dict(class_name=class_name, instance_name=instance_name)
        template_file_path = cast(str, os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "templates",
            os.path.splitext(os.path.basename(os.path.splitext(os.path.basename(__file__))[0]))[0],
            "task.py.jinja2"
        ))
        sys.exit(cls.render(template_file_path, render_data=render_data, render_file_path=render_file_path))



import math


def int_ceil(num: float) -> int:
    """
    >>> int_ceil(10 / 3)
    4

    :param num:
    :return:
    """
    return math.ceil(num)


def decimal_ceil(
        num: float,
        places: int = 2
) -> float:
    """
    >>> decimal_ceil(3.14159)
    3.15

    :param num:
    :param places:
    :return:
    """
    if places >= 0:
        scale = 10.0 ** places
        return math.ceil(num * scale) / scale
    else:
        raise ValueError(
            f"Invalid value for 'places': "
            f"Expected `int` and >= 0, "
            f"but got value: {places!r}"
        )


def format_number(
        number: int | float | str,
        places: int = 2
) -> str:
    """
    >>> format_number(3.1)
    '3.10'

    :param number:
    :param places:
    :return:
    """
    integer_str, decimal_str = str(float(number)).split(".")
    decimal_str = (decimal_str + "0" * (places - len(decimal_str)))[:places]
    number_str = ".".join([integer_str, decimal_str])
    return number_str



import subprocess
from typing import Any

import execjs
import py_mini_racer
import ujson


def _get_js_code(js_code: str | None = None, js_file_path: str | None = None) -> str:
    if js_code is None and js_file_path is None:
        raise ValueError(
            f"Either js_code: {js_code!r} or js_file_path: {js_file_path!r} must be provided"
        )
    if js_code is not None:
        return js_code
    if js_file_path is not None:
        with open(js_file_path, "r", encoding="utf-8") as f:
            js_code = f.read()
    return js_code


def execute_js_code_by_execjs(
        js_code: str | None = None,
        js_file_path: str | None = None,
        func_name: str | None = None,
        func_args: tuple[Any, ...] | None = None
) -> Any:
    js_code = _get_js_code(js_code, js_file_path)

    ctx = execjs.compile(js_code)
    if func_name is None:
        result = ctx.eval(js_code)
        return result
    if func_args is None:
        func_args = tuple()
    result = ctx.call(func_name, *func_args)
    return result


def execute_js_code_by_py_mini_racer(
        js_code: str | None = None,
        js_file_path: str | None = None,
        func_name: str | None = None,
        func_args: tuple[Any, ...] | None = None
) -> Any:
    js_code = _get_js_code(js_code, js_file_path)

    ctx = py_mini_racer.MiniRacer()
    result = ctx.eval(js_code)
    if func_name is None:
        return result
    if func_args is None:
        func_args = tuple()
    result = ctx.call(func_name, *func_args)
    return result


def execute_js_code_by_subprocess(
        js_code: str | None = None,
        js_file_path: str | None = None,
        func_args: tuple[dict[Any, Any], ...] | None = None,
) -> Any:
    js_code = _get_js_code(js_code, js_file_path)

    # js_code using `process.argv.slice(2)` to receive arguments
    func_args = ["node", "-e", js_code] + list(map(lambda x: ujson.dumps(x), func_args))
    process = subprocess.Popen(
        func_args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = process.communicate()
    result = stdout.decode()
    return result



from xproject.xdataclasses.xaccount_dataclass import AccountDataclass

from xproject.xdbs.xmongo_db import MongoDB
from xproject.xdbs.xmysql_db import MysqlDB
from xproject.xdbs.xredis_db import RedisDB

from xproject.xspider.xitems.xfield import Field
from xproject.xspider.xitems.xitem import Item
from xproject.xspider.xmodels.xmodel import Model
from xproject.xspider.xmodels.xsqlalchemy_model import SqlalchemyModel
from xproject.xspider.xenums.xdata_status_enum import DataStatusEnum

from xproject.xhandler import Handler
from xproject.xtask import Task

from xproject.xlogger import get_logger, logger

from xproject.xaliyun_oss import AliyunOSS
from xproject.xfrida import Frida
from xproject.xadb import ADB
from xproject.xwechat_ocr import WechatOCR

# script
import os

_assets_dir_name = "assets"  # noqa
ASSETS_DIR_PATH: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), _assets_dir_name)


def get_file_assets_dir_path(file_path: str, create_new_dir_path: bool = False) -> str:
    rel_path = os.path.relpath(os.path.abspath(file_path), start=ASSETS_DIR_PATH)
    new_dir_path = os.path.join(ASSETS_DIR_PATH, os.path.splitext(rel_path[3:])[0])  # 3 => ..\
    if create_new_dir_path:
        os.makedirs(new_dir_path, exist_ok=True)
    return new_dir_path


__all__ += ["get_file_assets_dir_path"]  # noqa
# script



import os
import time
from typing import Callable

from wechat_ocr.ocr_manager import OcrManager


class WechatOCR:
    def __init__(
            self,
            wechat_dir: str,
            wechat_ocr_file_path: str,
            ocr_callback: Callable[[str, dict], None] | None = None,
    ) -> None:
        self.wechat_dir = wechat_dir
        self.wechat_ocr_file_path = wechat_ocr_file_path
        self._default_ocr_callback_results: dict[str, dict] = {}

        def default_ocr_callback(image_file_path: str, results: dict) -> None:
            self._default_ocr_callback_results[image_file_path] = results

        self.ocr_callback = ocr_callback or default_ocr_callback

        self.ocr_manager = OcrManager(self.wechat_dir)
        self.ocr_manager.SetExePath(self.wechat_ocr_file_path)
        self.ocr_manager.SetUsrLibDir(self.wechat_dir)
        self.ocr_manager.SetOcrResultCallback(self.ocr_callback)

    def start(self) -> None:
        self.ocr_manager.StartWeChatOCR()

    def stop(self) -> None:
        self.ocr_manager.KillWeChatOCR()

    def ocr(
            self,
            image_file_paths: str | list[str],
            wait: float = 1.0
    ) -> dict[str, dict]:
        if isinstance(image_file_paths, str):
            image_file_paths = [image_file_paths]
        image_file_paths = list(set(image_file_paths))
        image_file_paths = [image_file_path for image_file_path in image_file_paths if os.path.exists(image_file_path)]

        self._default_ocr_callback_results.clear()

        for image_file_path in image_file_paths:
            self.ocr_manager.DoOCRTask(image_file_path)
            time.sleep(wait)

        while len(self._default_ocr_callback_results) < len(image_file_paths):
            time.sleep(0.1)

        return self._default_ocr_callback_results


if __name__ == "__main__":
    def main():
        wechat_dir = r"C:\Program Files\Tencent\WeChat\[3.9.12.55]"
        wechat_ocr_file_path = r"C:\Users\qst\AppData\Roaming\Tencent\WeChat\XPlugin\Plugins\WeChatOCR\7079\extracted\WeChatOCR.exe"

        wechat_ocr = WechatOCR(wechat_dir, wechat_ocr_file_path)
        wechat_ocr.start()

        from xproject import get_file_assets_dir_path
        print(wechat_ocr.ocr(os.path.join(get_file_assets_dir_path(__file__), "demo.png")))
        print(wechat_ocr.ocr(os.path.join(get_file_assets_dir_path(__file__), "demo.png")))

        wechat_ocr.stop()


    main()



from __future__ import annotations

import os
import random
import time
from dataclasses import dataclass
from functools import wraps
from typing import Callable, TypeVar, ParamSpec, Any, overload

import oss2


@dataclass
class OSSConfig:
    endpoint: str
    bucket_name: str
    access_key_id: str
    access_key_secret: str
    security_token: str | None = None

    @classmethod
    def from_env(cls) -> OSSConfig:
        return cls(
            endpoint=os.environ.get("ALIYUN_OSS_ENDPOINT", ""),
            bucket_name=os.environ.get("ALIYUN_OSS_BUCKET_NAME", ""),
            access_key_id=os.environ.get("ALIYUN_OSS_ACCESS_KEY_ID", ""),
            access_key_secret=os.environ.get("ALIYUN_OSS_ACCESS_KEY_SECRET", ""),
            security_token=os.environ.get("ALIYUN_OSS_SECURITY_TOKEN"),
        )


def _is_retryable(exception: Exception) -> bool:
    if isinstance(exception, oss2.exceptions.RequestError):
        return True
    return False


P = ParamSpec("P")
R = TypeVar("R")

Decorator = Callable[[Callable[P, R]], Callable[P, R]]


@overload
def retry(
        func: None = ...,
        *,
        max_retries: int = ...,
        delay: float = ...,
        backoff: float = ...,
        jitter: float = ...,
) -> Decorator:
    ...


@overload
def retry(
        func: Callable[P, R],
        *,
        max_retries: int = ...,
        delay: float = ...,
        backoff: float = ...,
        jitter: float = ...,
) -> Callable[P, R]:
    ...


def retry(
        func: Callable[P, R] | None = None,
        *,
        max_retries: int = 3,
        delay: float = 0.5,
        backoff: float = 2.0,
        jitter: float = 0.1,
) -> Any:
    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        @wraps(fn)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            delay_ = max(0.0, delay)
            last_exception: Exception | None = None

            for retries in range(1, max(1, max_retries) + 1):
                try:
                    return fn(*args, **kwargs)
                except Exception as exception:
                    last_exception = exception
                    if not _is_retryable(exception) or retries >= max_retries:
                        raise
                    time.sleep(delay_ + random.uniform(0.0, jitter))
                    delay_ *= max(1.0, backoff)

            raise last_exception.with_traceback(last_exception.__traceback__)

        return wrapper

    if func is None:
        return decorator

    return decorator(func)


class AliyunOSS:
    def __init__(self, bucket: oss2.Bucket):
        self._bucket = bucket

    @classmethod
    def from_config(cls, config: OSSConfig) -> AliyunOSS:
        auth = (
            oss2.Auth(config.access_key_id, config.access_key_secret)
            if not config.security_token
            else oss2.StsAuth(config.access_key_id, config.access_key_secret, config.security_token)
        )
        bucket = oss2.Bucket(auth, config.endpoint, config.bucket_name)
        return cls(bucket)

    @classmethod
    def from_params(
            cls,
            endpoint: str,
            bucket_name: str,
            access_key_id: str,
            access_key_secret: str,
            security_token: str | None = None,
    ) -> AliyunOSS:
        return cls.from_config(OSSConfig(endpoint, bucket_name, access_key_id, access_key_secret, security_token))

    # ========== upload ==========
    @retry
    def put_content(
            self,
            key: str,
            content: bytes,
            *,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None,
            progress_callback: Callable[[int, int | None], None] | None = None
    ) -> oss2.models.PutObjectResult:
        return self._bucket.put_object(key, content, headers=headers, progress_callback=progress_callback)

    @retry
    def put_text(
            self,
            key: str,
            text: str,
            *,
            encoding: str = "utf-8",
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None,
            progress_callback: Callable[[int, int | None], None] | None = None
    ) -> oss2.models.PutObjectResult:
        data = text.encode(encoding)
        return self._bucket.put_object(key, data, headers=headers, progress_callback=progress_callback)

    @retry
    def put_object_from_file(
            self,
            key: str,
            filename: str,
            *,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None,
            progress_callback: Callable[[int, int | None], None] | None = None
    ) -> oss2.models.PutObjectResult:
        return self._bucket.put_object_from_file(key, filename, headers=headers, progress_callback=progress_callback)

    # ========== download ==========
    @retry
    def get_bytes(
            self,
            key: str,
            *,
            byte_range: tuple[int | None, int | None] | None = None,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None,
            progress_callback: Callable[[int, int | None], None] | None = None,
            process: str | None = None,
            params: dict[str, str] | None = None
    ) -> bytes:
        response = self._bucket.get_object(
            key,
            byte_range=byte_range,
            headers=headers,
            progress_callback=progress_callback,
            process=process,
            params=params
        )
        try:
            return response.read()
        finally:
            response.close()

    @retry
    def get_text(
            self,
            key: str,
            *,
            encoding: str = "utf-8",
            byte_range: tuple[int | None, int | None] | None = None,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None,
            progress_callback: Callable[[int, int | None], None] | None = None,
            process: str | None = None,
            params: dict[str, str] | None = None
    ) -> str:
        response = self._bucket.get_object(
            key,
            byte_range=byte_range,
            headers=headers,
            progress_callback=progress_callback,
            process=process,
            params=params
        )
        try:
            return response.read().decode(encoding)
        finally:
            response.close()

    # ========== manage ==========
    @retry
    def object_exists(
            self,
            key: str,
            *,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None
    ) -> bool:
        return self._bucket.object_exists(key, headers=headers)

    @retry
    def head_object(
            self,
            key: str,
            *,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None,
            params: dict[str, str] | oss2.CaseInsensitiveDict | None = None
    ) -> oss2.models.HeadObjectResult:
        return self._bucket.head_object(key, headers=headers, params=params)

    @retry
    def delete_object(
            self,
            key: str,
            *,
            params: dict[str, str] | None = None,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None
    ) -> oss2.models.RequestResult:
        return self._bucket.delete_object(key, params=params, headers=headers)

    @retry
    def batch_delete_objects(
            self,
            key_list: list[str],
            *,
            headers: dict[str, str] | oss2.CaseInsensitiveDict | None = None
    ) -> oss2.models.BatchDeleteObjectsResult | None:
        if not key_list:
            return None
        return self._bucket.batch_delete_objects(key_list, headers=headers)



import os
from typing import Any

import jinja2

from xproject.xstring import camel_to_snake, snake_to_camel


class Render:
    @classmethod
    def render(
            cls,
            template_file_path: str | None = None,
            template_dir_path: str | None = None, template_file_name: str | None = None,
            render_data: dict[str, Any] | None = None,
            render_file_path: str | None = None
    ) -> str:
        if template_file_path is None and (template_file_name is None or template_dir_path is None):
            raise ValueError(
                f"Either template_file_path: {template_file_path!r} or "
                f"template_file_name: {template_file_name!r} and template_dir_path: {template_dir_path!r} "
                f"must be provided"
            )

        if template_file_path is not None:
            template_file_path = os.path.abspath(template_file_path)

        if template_dir_path is not None:
            template_dir_path = os.path.abspath(template_dir_path)

        if render_data is None:
            render_data = dict()

        if template_file_path is not None:
            template_dir_path = os.path.dirname(template_file_path)
            template_file_name = os.path.basename(template_file_path)

        loader = jinja2.FileSystemLoader(searchpath=template_dir_path)
        env = jinja2.Environment(loader=loader)
        env.filters["camel_to_snake"] = camel_to_snake
        env.filters["snake_to_camel"] = snake_to_camel
        template = env.get_template(name=template_file_name)
        result = template.render(**render_data)

        if render_file_path is not None:
            with open(render_file_path, "w", encoding="utf-8") as file:
                file.write(result)

        return result



from abc import ABCMeta
from collections.abc import ItemsView, KeysView, ValuesView
from types import new_class
from typing import NamedTuple, cast

from xproject.xstring import camel_to_snake


class ViewClasses(NamedTuple):
    ItemsView: type
    KeysView: type
    ValuesView: type


def create_view_classes(class_name: str) -> ViewClasses:
    def _create_view_classes(base_class: type) -> type:
        name = f"{class_name}{base_class.__name__}"
        return new_class(
            name,
            (base_class,),
            exec_body=lambda ns: ns.update({
                "__repr__": lambda self: f"{camel_to_snake(name)}({list(self)})"
            })
        )

    return ViewClasses(
        ItemsView=_create_view_classes(ItemsView),
        KeysView=_create_view_classes(KeysView),
        ValuesView=_create_view_classes(ValuesView),
    )


def create_subclasscheck_meta_class(
        required_all_methods: tuple[str, ...],
        required_any_methods: tuple[str, ...] = tuple(),
        meta_class_name: str = "SubclasscheckMeta"
) -> type[ABCMeta]:
    return cast(type[ABCMeta], new_class(
        meta_class_name,
        (ABCMeta,),
        exec_body=lambda ns: ns.update({
            "__subclasscheck__": classmethod(
                lambda cls, subclass:
                all([callable(getattr(subclass, i, None)) for i in required_all_methods]) and
                any([callable(getattr(subclass, i, None)) for i in required_any_methods]) if required_any_methods else
                True
            )
        })
    ))


__all__ = [
    "ViewClasses", "create_view_classes", "create_subclasscheck_meta_class"
]



from typing import Any, TypeGuard


def is_bool(obj: Any) -> TypeGuard[bool]:
    return isinstance(obj, bool)


def is_int(obj: Any) -> TypeGuard[int]:
    return isinstance(obj, int) and not isinstance(obj, bool)


def is_str(val: Any) -> TypeGuard[str]:
    return isinstance(val, str)


def is_list(val: Any) -> TypeGuard[list[Any]]:
    return isinstance(val, list)


def is_list_of[T](val: list[Any], ele_type: type[T]) -> TypeGuard[list[T]]:
    return isinstance(val, list) and all(isinstance(i, ele_type) for i in val)


__all__ = [
    "is_bool", "is_int", "is_str", "is_list", "is_list_of",
]



from importlib import import_module
from inspect import isclass

from xproject.xexceptions import ImportClassTypeError
from xproject.xvalidators import is_str


def import_class(class_ref: str | type) -> type:
    if is_str(class_ref):
        module_path, class_name = class_ref.rsplit(".", maxsplit=1)
        module = import_module(module_path)
        cls = getattr(module, class_name, None)
    elif isclass(class_ref):
        cls = class_ref
    else:
        raise ImportClassTypeError(
            f"Invalid type for 'class_ref': "
            f"Expected `str | type`, "
            f"but got {type(class_ref).__name__!r} (value: {class_ref!r})"
        )
    return cls



from __future__ import annotations

import re
from abc import ABCMeta
from collections.abc import MutableMapping, Iterator
from copy import deepcopy
from pprint import pformat
from types import new_class
from typing import Any, Self, Final

from scrapy.item import Field as ScrapyField, Item as ScrapyItem

from xproject.xexceptions import InitException, GetattributeException
from xproject.xspider.xitems.xfield import Field
from xproject.xspider.xmodels.xmodel import Model
from xproject.xtypes import ViewClasses, create_view_classes


class ScrapyItemImpl(ScrapyItem):
    def to_xproject_item(self) -> Item: ...


class ItemMeta(ABCMeta):
    def __new__(mcs, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> type:
        fields = dict()
        new_attrs = dict()

        if (model_class := attrs.get("MODEL")) is not None:
            model_class: type[Model]
            for column in model_class.columns():
                fields[column] = Field()

        for key, value in attrs.items():
            if isinstance(value, Field):
                fields[key] = value
            else:
                new_attrs[key] = value

        cls = super().__new__(mcs, name, bases, new_attrs)

        cls._FIELDS = fields

        cls._SCRAPY_ITEM = new_class(
            re.sub(r"Item$", "", name) + "ScrapyItem", (ScrapyItem,),
            exec_body=lambda ns: ns.update(
                {field: ScrapyField() for field in cls._FIELDS} |
                {"to_xproject_item": lambda self: cls.create_by_scrapy_item(self)}  # noqa
            )
        )

        return cls


class Item(MutableMapping, metaclass=ItemMeta):
    MODEL: type[Model]
    _FIELDS: dict[str, Field]
    _SCRAPY_ITEM: type[ScrapyItemImpl]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._item: Final[dict[str, Any]] = dict()

        if args:
            raise InitException(
                f"{self.__class__.__name__} initialization failed: "
                "Positional arguments are not supported "
                "(use keyword arguments instead)"
            )
        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self._FIELDS:
            self._item[key] = value
        else:
            raise KeyError(
                f"{self.__class__.__name__} field assignment error: "
                f"Field {key!r} is not defined "
                f"(add the field to class definition first)"
            )

    def __delitem__(self, key: str) -> None:
        del self._item[key]

    def __getitem__(self, key: str) -> Any:
        return self._item[key]

    def __setattr__(self, name: str, value: Any) -> None:
        if name != "_item":
            raise AttributeError(
                f"{self.__class__.__name__} attribute assignment error: "
                f"Cannot set attribute {name!r} directly "
                f"(to set a field value, use item[{name!r}] = {value!r} instead)"
            )
        super().__setattr__(name, value)

    def __getattribute__(self, name: str) -> Any:
        fields = super().__getattribute__("_FIELDS")
        if name in fields:
            raise GetattributeException(
                f"{self.__class__.__name__} attribute access error: "
                f"Cannot get attribute {name!r} directly "
                f"(to access a field value, use item[{name!r}] instead)"
            )
        return super().__getattribute__(name)

    def __getattr__(self, name: str) -> None:
        raise AttributeError(
            f"{self.__class__.__name__} attribute access error: "
            f"Attribute {name!r} is not defined "
            f"(to access a field value, first add the field to class definition, then use item[{name!r}])"
        )

    def __iter__(self) -> Iterator[str]:
        return iter(self._item)

    def __len__(self) -> int:
        return len(self._item)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}\n{pformat(self._item)}\n>"

    __str__ = __repr__

    @classmethod
    def get_model_class(cls) -> type[Model]:
        return cls.MODEL

    @classmethod
    def get_scrapy_item_class(cls) -> type[ScrapyItemImpl]:
        return cls._SCRAPY_ITEM

    @classmethod
    def create_scrapy_item(cls, data: dict[str, Any] | None = None) -> ScrapyItemImpl:
        if data is not None:
            return cls.get_scrapy_item_class()(**data)
        return cls.get_scrapy_item_class()()

    @classmethod
    def create_by_scrapy_item(cls, scrapy_item: ScrapyItem) -> Self:
        return cls(**dict(scrapy_item))

    def to_dict(self) -> dict[str, Any]:
        return dict(self)

    def to_scrapy_item(self) -> ScrapyItemImpl:
        return self.create_scrapy_item(self.to_dict())

    def copy(self) -> Self:
        return deepcopy(self)

    def save(self) -> bool:
        return self.get_model_class().save(self)

    @classmethod
    @property
    def view_classes(cls) -> ViewClasses:  # noqa
        return create_view_classes(cls.__name__)

    def items(self) -> ViewClasses.ItemsView:
        return self.view_classes.ItemsView(self)

    def keys(self) -> ViewClasses.KeysView:
        return self.view_classes.KeysView(self)

    def values(self) -> ViewClasses.ValuesView:
        return self.view_classes.ValuesView(self)

    @property
    def unassigned_keys(self) -> list[str]:
        return [k for k in self._FIELDS.keys() if k not in self._item.keys()]


__all__ = [
    "Item"
]

if __name__ == '__main__':
    def main():
        class TestItem(Item):
            name = Field()
            age = Field()

        try:
            item = TestItem("xspider", 18)
        except Exception as e:
            print(type(e), e)

        try:
            item = TestItem()
            item["gender"] = "man"
        except Exception as e:
            print(type(e), e)

        try:
            item = TestItem()
            item.gender = "man"
        except Exception as e:
            print(type(e), e)

        try:
            item = TestItem()
            print(item.name)
        except Exception as e:
            print(type(e), e)

        try:
            item = TestItem()
            print(item.gender)
        except Exception as e:
            print(type(e), e)

        item = TestItem()
        item["name"] = "xproject"
        item["age"] = 18
        print(item)

        print(item.items())


    main()



import pandas as pd


def split_df(df: pd.DataFrame, num_parts: int | None = None, part_size: int | None = None) -> list[pd.DataFrame]:
    if df.empty:
        return []

    if num_parts:
        avg_size = len(df) // num_parts
        remainder = len(df) % num_parts

        result = []
        start = 0
        for i in range(num_parts):
            end = start + avg_size + (1 if i < remainder else 0)
            result.append(df.iloc[start:end])
            start = end
        return result

    elif part_size:
        return [df.iloc[i:i + part_size] for i in range(0, len(df), part_size)]

    else:
        raise ValueError(
            f"Either num_parts: {num_parts!r} or part_size: {part_size!r} must be provided"
        )


if __name__ == '__main__':
    def main():
        df = pd.DataFrame({"A": range(1, 101), "B": range(101, 201)})
        # dfs = split_df(df, num_parts=2)
        dfs = split_df(df, part_size=50)
        for df in dfs:
            print(df)


    main()



import random
import threading

import requests

from xproject.xlogger import get_logger


class Clash:
    def __init__(
            self,
            api_url: str = "http://127.0.0.1:9097",
            proxy_group: str = "GLOBAL",
            proxy_port: int = 7897,
            api_secret: str = "set-your-secret",
            switch_interval: int = 60,
    ):
        self.api_url = api_url.rstrip("/")
        self.proxy_group = proxy_group
        self.proxy_port = proxy_port
        self.api_secret = api_secret
        self.switch_interval = switch_interval

        self._running_event = threading.Event()
        self._thread: threading.Thread | None = None
        self.logger = get_logger()

    def _headers(self) -> dict:
        return {"Authorization": f"Bearer {self.api_secret}"} if self.api_secret else {}

    def _get(self, path: str) -> requests.Response:
        url = f"{self.api_url}/{path.lstrip('/')}"
        return requests.get(url, headers=self._headers(), timeout=5)

    def _put(self, path: str, data: dict) -> requests.Response:
        url = f"{self.api_url}/{path.lstrip('/')}"
        return requests.put(url, json=data, headers=self._headers(), timeout=5)

    def get_node_names(self) -> list[str]:
        response = self._get("/proxies")
        response.raise_for_status()
        jsn = response.json()
        return jsn.get("proxies", {}).get(self.proxy_group, {}).get("all", [])

    def switch_to_node(self, node_name: str) -> str | None:
        response = self._put(f"/proxies/{self.proxy_group}", {"name": node_name})
        if response.status_code == 204:
            return node_name
        return None

    def switch_to_random_node(self) -> str | None:
        node_names = self.get_node_names()
        if node_names:
            node_name = random.choice(node_names)
            node_name = self.switch_to_node(node_name)
            return node_name
        return None

    def get_current_node_name(self) -> str:
        response = self._get(f"/proxies/{self.proxy_group}")
        response.raise_for_status()
        return response.json().get("now", "")

    def get_current_ip(self) -> str:
        proxies = {
            "http": f"http://127.0.0.1:{self.proxy_port}",
            "https": f"http://127.0.0.1:{self.proxy_port}",
        }
        response = requests.get("https://ipinfo.io/ip", proxies=proxies, timeout=8)
        return response.text.strip()

    def _run_service(self) -> None:
        while self._running_event.is_set():
            try:
                self.switch_to_random_node()

                node_name = self.get_current_node_name()
                if node_name:
                    self.logger.debug(f"node_name: {node_name}")

                ip = self.get_current_ip()
                if ip:
                    self.logger.debug(f"ip: {ip}")

            except Exception as e:
                self.logger.error(f"Exception: {e}")

            if not self._running_event.wait(timeout=self.switch_interval):
                break

    def start_service(self) -> None:
        if not self._running_event.is_set():
            self._running_event.set()
            self._thread = threading.Thread(target=self._run_service, daemon=True)
            self._thread.start()
            self.logger.info("started service")

    def stop_service(self) -> None:
        if self._running_event.is_set():
            self._running_event.clear()
            self.logger.debug("send stop service signal...")

            if self._thread and self._thread.is_alive():
                self._thread.join()
                self.logger.info("stoped service")

    def listen(self) -> None:
        try:
            while self._running_event.is_set():
                threading.Event().wait(1)
        except KeyboardInterrupt:
            self.logger.warning("KeyboardInterrupt received, stopping service...")
            self.stop_service()


if __name__ == "__main__":
    clash = Clash(switch_interval=10)
    clash.start_service()
    clash.listen()
    # clash.stop_service()



from . import xasync_context_manager_object_mixin
from . import xasync_create_instance_object_mixin
from . import xcontext_manager_object_mixin
from . import xcreate_instance_object_mixin
from . import xobject_mixin

__all__ = [
    "xasync_context_manager_object_mixin",
    "xasync_create_instance_object_mixin",
    "xcontext_manager_object_mixin",
    "xcreate_instance_object_mixin",
    "xobject_mixin",
]
